#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 24 09:30:15 2023

@author: Dr. Lauren J Beesley VanDervort, lvandervort@lanl.gov
"""

### Imports
from scipy.optimize import basinhopping
from scipy.optimize import minimize
import gc
import pandas as pd
import numpy as np
import seaborn as sns
import scipy.stats
from impala import superCal as sc
from impala import physics as pm_vec
from scipy.optimize import NonlinearConstraint
from scipy.stats import halfcauchy
import sklearn
import pyswarm
from scipy.stats import qmc
import time
import os
import multiprocess as mp


### Function for obtaining the MAP estimator associated
### with the pooled Impala model
def get_map_impalapool(setup, n_samples = 1000, theta_init = None, disc_init = None, optmethod = 'bh', niter = None, T = 1, n_cores = 10): 
    
    if str(type(niter)) == "<class 'NoneType'>":
        niter = setup.p*10

    ### Draw s2
    s2 = [np.empty([n_samples,len(setup.ig_a[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        if setup.models[i] == 'fix':
            s2[i] = setup.sd_est[i]
        else:
            for j in range(len(setup.ig_a[i])):
                if setup.s2_df[i][j] == 0:
                    s2[i][:,j] = halfcauchy.rvs(loc = 0, scale = 1, size = n_samples).reshape(1,n_samples)    
                else:
                    s2[i][:,j] = invgamma.rvs(a=setup.ig_a[i][j], scale = setup.ig_b[i][j], size = n_samples).reshape(1,n_samples)
    s2_expand = [np.empty([n_samples,len(setup.ys[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        for j in range(len(setup.ig_a[i])):
            s2_expand[i][:,setup.s2_ind[i] == j] = s2[i][:,j].reshape(n_samples,1)
    
    ### Handle constraints
    if optmethod == 'bh':
        def my_constraint(y):
            y = np.reshape(y, (1, -1))
            A = int(setup.checkConstraints(sc.tran_unif(y[0,0:setup.p].reshape(1,-1),setup.bounds_mat, setup.bounds.keys()), setup.bounds)[0])
            return A    
    else:
        def my_constraint(y):
            y = np.reshape(y, (1, -1))
            A = int(setup.checkConstraints(sc.tran_unif(y[0,0:setup.p].reshape(1,-1),setup.bounds_mat, setup.bounds.keys()), setup.bounds)[0])
            return (A-0.5)
    non_linear_constr = NonlinearConstraint(my_constraint, 0.5, np.inf)
    ### Optimize
    def neg_log_lik(y):
        y = np.reshape(y, (1, -1))
        if y.shape[1] > setup.p:
            v_cur = y[0,setup.p:y.shape[1]]
            disc_dims = [setup.models[i].nd for i in range(setup.nexp)]
            v_cur_list = [v_cur[0:disc_dims[i]] if i == 0 else v_cur[int(np.sum(disc_dims[0:(i-1)])):int(np.sum(disc_dims[0:i]))] for i in range(setup.nexp)]
            theta_cur = y[0,0:setup.p].reshape(1,-1)
        else:
            theta_cur = y[0,:].reshape(1,-1)
        CONSTRAINTS = setup.checkConstraints(sc.tran_unif(theta_cur,setup.bounds_mat, setup.bounds.keys()), setup.bounds)
        if CONSTRAINTS[0] == True and (theta_cur<=0).sum() == 0 and (theta_cur>=1).sum() == 0:
            disc_y = [(setup.models[i].D @ v_cur_list[i]).reshape(1,-1) if setup.models[i].nd > 0 else np.repeat(0,len(setup.ys[i])).reshape(1,-1) for i in range(setup.nexp)]     
            pred_y = [setup.models[i].eval(sc.tran_unif(theta_cur,setup.bounds_mat, setup.bounds.keys()), pool=True) for i in range(setup.nexp)]
            loglik_y = [-0.5* (((((setup.ys[i] - pred_y[i] - disc_y[i])**2)/s2_expand[i]) +np.log(s2_expand[i])).sum(axis = 1)).reshape(n_samples,1) for i in range(setup.nexp)]
            loglik_y_disc = [-0.5* (((v_cur_list[i]**2)/setup.models[i].discrep_tau) +np.log(setup.models[i].discrep_tau)).sum() if setup.models[i].nd>0 else 0 for i in range(setup.nexp)]
            llik = np.nansum(loglik_y,axis=0)+np.nansum(np.hstack(loglik_y_disc))
            llik_max = llik.max()
            llik = llik - llik_max  
            llik[llik < -10**2] = -10**2 #avoid exp underflow issues
            output = np.log(np.exp(llik).mean()) + llik_max
            output = np.array(output)
            output = output.reshape(1,1)
            output = -1*output[0][0]
        else:
            dist_func = getattr(setup, 'distConstraints', None)
            if str(type(dist_func)) != "<class 'NoneType'>":
                VAL = setup.distConstraints(sc.tran_unif(y,setup.bounds_mat, setup.bounds.keys()), setup.bounds)
                output = (10 ** 6) + np.sum(VAL)
            else:
                output = 10 ** 6
        return output
        
        
    if str(type(disc_init)) != "<class 'NoneType'>":
        param_cur = np.append(sc.normalize(theta_init.flatten(),setup.bounds_mat),np.hstack(disc_init))
    else:
        param_cur = sc.normalize(theta_init.flatten(),setup.bounds_mat)

    if optmethod == 'bh':
        res = basinhopping(neg_log_lik, param_cur, minimizer_kwargs={"method":"COBYLA","constraints":non_linear_constr}, niter=niter, disp = True, stepsize = 0.1, T = T)
    elif optmethod == 'pso':
        res = pyswarm.pso(neg_log_lik, lb = np.repeat(0,len(param_cur.flatten())), ub = np.repeat(1,len(param_cur.flatten())), ieqcons=[my_constraint], swarmsize=niter, debug = True, omega = 0.9, maxiter = niter)
    elif optmethod == 'grid':
        from scipy.stats import qmc
        import time
        import os
        import multiprocess as mp
        from scipy.optimize import minimize
        sampler = qmc.LatinHypercube(d=setup.p)
        sample = sampler.random(n=10000)  
        sample = np.append(sc.normalize(theta_init.flatten(),setup.bounds_mat).reshape(1,-1), sample,axis=0)
        CONSTRAINTS = setup.checkConstraints(sc.tran_unif(sample,setup.bounds_mat, setup.bounds.keys()), setup.bounds)
        sample = sample[CONSTRAINTS,:]
        niter = np.min((niter,sample.shape[0]))
        sample = sample[0:niter,:]
        
        if str(type(disc_init)) != "<class 'NoneType'>":
            def run_mcmc_in_parallel(i):
                disc_rand = np.hstack(disc_init) + np.random.normal(loc = np.repeat(0,len(disc_init.flatten())),scale =np.sqrt(np.repeat(setup.models[0].discrep_tau,len(disc_init.flatten()))), size = len(disc_init.flatten()))
                res = minimize(neg_log_lik, np.append(sample[i,:].reshape(1,-1), disc_rand.reshape(1,-1),axis=1).flatten(), method = 'COBYLA', constraints = non_linear_constr)
                return (res.x)
        else:
            def run_mcmc_in_parallel(i):
                res = minimize(neg_log_lik, sample[i,:], method = 'COBYLA', constraints = non_linear_constr)
                return (res.x)            
        with mp.Pool(n_cores) as pool:
            A = pool.map(run_mcmc_in_parallel, range(niter))
    else:
        print('Invalid optmethod')
        return
    
    import copy
    if optmethod == 'bh':
        res_trans0 = copy.deepcopy(res.x)
    elif optmethod == 'pso':
        res_trans0 = copy.deepcopy(res[0])
    elif optmethod == 'grid':
        B = np.apply_along_axis(neg_log_lik, 1, A)
        res_trans0 = A[np.where(B == B.min())[0][0]]
    res_trans = sc.tran_unif(res_trans0[0:setup.p],setup.bounds_mat, setup.bounds.keys())
    
    if len(res_trans0) > setup.p:
        res_trans['v'] = res_trans0[setup.p:len(res_trans0)]
    return res_trans



### Monte Carlo integrated posterior for pooled impala model, integrating out everything but v (basis coefficients for discrepancy)
def eval_partialintlogposterior_impalapool(setup, n_samples = 1000, theta = None, disc_v = None): 
    ### Draw s2
    s2 = [np.empty([n_samples,len(setup.ig_a[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        if setup.models[i].s2 == 'fix':
            for j in range(len(setup.ig_a[i])):
                s2[i][:,j] = np.repeat(setup.sd_est[i],s2[i].shape[0])
        else:
            for j in range(len(setup.ig_a[i])):
                if setup.s2_df[i][j] == 0:
                    s2[i][:,j] = halfcauchy.rvs(loc = 0, scale = 1, size = n_samples).reshape(1,n_samples)    
                else:
                    s2[i][:,j] = invgamma.rvs(a=setup.ig_a[i][j], scale = setup.ig_b[i][j], size = n_samples).reshape(1,n_samples)
    s2_expand = [np.empty([n_samples,len(setup.ys[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        for j in range(len(setup.ig_a[i])):
            s2_expand[i][:,setup.s2_ind[i] == j] = s2[i][:,j].reshape(n_samples,1)
    
    ### Optimize
    def neg_log_lik(y):
        y = np.reshape(y, (1, -1))
        if y.shape[1] > setup.p:
            v_cur = y[0,setup.p:y.shape[1]]
            disc_dims = [setup.models[i].nd for i in range(setup.nexp)]
            v_cur_list = [v_cur[0:disc_dims[i]] if i == 0 else v_cur[int(np.sum(disc_dims[0:(i-1)])):int(np.sum(disc_dims[0:i]))] for i in range(setup.nexp)]
            theta_cur = y[0,0:setup.p].reshape(1,-1)
        else:
            theta_cur = y[0,:].reshape(1,-1)
        pred_y = [np.empty([n_samples,len(setup.ys[i])]) for i in range(setup.nexp)]
        for i in range(setup.nexp):
            pred_y[i] = setup.models[i].eval(sc.tran_unif(theta_cur,setup.bounds_mat, setup.bounds.keys()), pool=True)
        disc_y = [(setup.models[i].D @ v_cur_list[i]).reshape(1,-1) if setup.models[i].nd > 0 else np.repeat(0,len(setup.ys[i])).reshape(1,-1) for i in range(setup.nexp)]     
        loglik_y = [np.empty([n_samples,1]) for i in range(setup.nexp)]
        for i in range(setup.nexp):
            loglik_y[i] = -0.5* (((((setup.ys[i] - pred_y[i] - disc_y[i])**2)/s2_expand[i]) +np.log(s2_expand[i])).sum(axis = 1)).reshape(n_samples,1)
        llik = np.nansum(loglik_y,axis = 0) #loglik_y[0]
        #for i in range(setup.nexp-1):
        #    llik = llik + loglik_y[i+1]
        llik_max = llik.max()
        llik = llik - llik_max  
        llik[llik< -10**2]=-10**2 #preventing exp underflow issue
        output = np.log(np.exp(llik).mean()) + llik_max
        output = np.array(output)
        output = output.reshape(1,1)
        output = -1*output[0][0]
        return output
    if str(type(disc_v)) != "<class 'NoneType'>":
        loglik=-1*np.apply_along_axis(neg_log_lik, 1, np.append(sc.normalize(theta,setup.bounds_mat),disc_v,axis=1))
    else:
        loglik=-1*np.apply_along_axis(neg_log_lik, 1, sc.normalize(theta,setup.bounds_mat))
    return loglik


### Monte Carlo integrated posterior for pooled impala model, integrating out everything including v (basis coefficients for discrepancy)
def eval_fullintlogposterior_impalapool(setup, n_samples = 1000, theta = None): 
    ### Draw s2
    s2 = [np.empty([n_samples,len(setup.ig_a[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        if setup.models[i] == 'fix':
            for j in range(len(setup.ig_a[i])):
                s2[i][:,j] = np.repeat(setup.sd_est[i],s2[i].shape[0])
        else:
            for j in range(len(setup.ig_a[i])):
                if setup.s2_df[i][j] == 0:
                    s2[i][:,j] = halfcauchy.rvs(loc = 0, scale = 1, size = n_samples).reshape(1,n_samples)    
                else:
                    s2[i][:,j] = invgamma.rvs(a=setup.ig_a[i][j], scale = setup.ig_b[i][j], size = n_samples).reshape(1,n_samples)
    s2_expand = [np.empty([n_samples,len(setup.ys[i])]) for i in range(setup.nexp)]
    for i in range(setup.nexp):
        for j in range(len(setup.ig_a[i])):
            s2_expand[i][:,setup.s2_ind[i] == j] = s2[i][:,j].reshape(n_samples,1)
    
    disc_dims = [setup.models[i].nd for i in range(setup.nexp)]
    
    ### Draw v_cur
    v_cur = np.empty([n_samples,np.sum(disc_dims)])
    for i in range(setup.nexp):
        if 'discrep_tau' not in dir(setup.models[i]):
            setup.models[i].discrep_tau = 1e-10
        if i == 0:
            v_cur[:,0:disc_dims[i]] = np.random.normal(size = n_samples * disc_dims[i], loc = 0, scale = np.sqrt(setup.models[i].discrep_tau)).reshape(n_samples, disc_dims[i])
        else:
            v_cur[:,np.sum(disc_dims[0:i]):np.sum(disc_dims[0:(i+1)])] = np.random.normal(size = n_samples * disc_dims[i], loc = 0, scale = np.sqrt(setup.models[i].discrep_tau)).reshape(n_samples, disc_dims[i])
    v_cur_list = [v_cur[:,0:disc_dims[i]] if i == 0 else v_cur[:,int(np.sum(disc_dims[0:(i-1)])):int(np.sum(disc_dims[0:i]))] for i in range(setup.nexp)]
  
    ### Optimize
    def neg_log_lik(y):
        y = np.reshape(y, (1, -1))
        pred_y = [np.empty([n_samples,len(setup.ys[i])]) for i in range(setup.nexp)]
        for i in range(setup.nexp):
            pred_y[i] = setup.models[i].eval(sc.tran_unif(y,setup.bounds_mat, setup.bounds.keys()), pool=True)
        disc_y = [(setup.models[i].D @ v_cur_list[i].T).T if setup.models[i].nd > 0 else np.repeat(0,len(setup.ys[i])).reshape(n_samples,-1) for i in range(setup.nexp)]     
        loglik_y = [np.empty([n_samples,1]) for i in range(setup.nexp)]
        for i in range(setup.nexp):
            loglik_y[i] = -0.5* (((((setup.ys[i] - pred_y[i] - disc_y[i])**2)/s2_expand[i]) +np.log(s2_expand[i])).sum(axis = 1)).reshape(n_samples,1)
        llik = loglik_y[0]
        for i in range(setup.nexp-1):
            llik = llik + loglik_y[i+1]
        llik_max = llik.max()
        llik = llik - llik_max  
        llik[llik < -10**2] = -10**2
        output = np.log(np.exp(llik).mean()) + llik_max
        output = np.array(output)
        output = output.reshape(1,1)
        output = -1*output[0][0]
        return output
    loglik=-1*np.apply_along_axis(neg_log_lik, 1, sc.normalize(theta,setup.bounds_mat))
    return loglik




import statistics as stat
import unittest

import numpy as np

from impala.physics import physical_models_vec as pmh


# Constant yield stress model
class TestJCYieldStress(unittest.TestCase):
    def setUp(self):
        """
        Parameters are for OFHC copper.
        """
        self.params = {
            "A": 0.00090,  # MBar
            "B": 0.00292,  # MBar
            "C": 0.0250,  # -
            "n": np.array([0.31]),  # -
            "m": np.array([1.09]),  # -
        }
        self.consts = {
            "rho0": 8.9375,  # g/cc
            "Cv0": 0.383e-5,  # MBar*cc/g*K
            "G0": 0.4578,  # MBar
            "Tmelt0": 1356.0,  # K
            "Tref": 298.0,  # K
            "edot0": 1.0e-6,  # 1/micro-s
            "chi": 1.0,
        }
        self.model_yjc = pmh.MaterialModel(
            flow_stress_model=pmh.JC_Yield_Stress,
        )

    def test_isothermal_lowrate(self):
        """
        JC, const. g0, Tm. Rates less than 1e-6/us (~1/s), should be isothermal T=const
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_yjc.initialize(self.params, self.consts)
        self.model_yjc.initialize_state(T=np.array([298.0]))
        results_yjc = self.model_yjc.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertEqual(results_yjc[0][3], results_yjc[-1][3])

    def test_adiabatic_highrate(self):
        """
        JC, const. g0, Tm. Rates greater than 1e-6/us (~1/s), temperature changes adiabatically
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_yjc.initialize(self.params, self.consts)
        self.model_yjc.initialize_state(T=np.array([298.0]))
        results_yjc = self.model_yjc.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertNotEqual(results_yjc[0][3], results_yjc[-1][3])

    def test_stress_highrate(self):
        """
        JC, constant g0 and Tm, high rate 1e4/s, FLAG/python comparison within 2%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-2(micro-s)^-1 = 1e4(s)^-1

        FLAG sim results
        eps_p = 0.28179, flow_stress =  343.37 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.28179]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_yjc.initialize(self.params, self.consts)
        self.model_yjc.initialize_state(T=np.array([298.0]))
        results_yjc = self.model_yjc.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_yjc[-1, 2, 0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.02
        stress_FLAG = 343.37e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / stat.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midrate(self):
        """
        JC, constant g0 and Tm, mid rate 1e2/s, FLAG/python comparison within 1%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-4(micro-s)^-1 = 1e2(s)^-1

        FLAG sim results
        eps_p = 0.28895, flow_stress =  317.06 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.28895]), edot=np.array([1e-4]), nhist=1000
        )
        self.model_yjc.initialize(self.params, self.consts)
        self.model_yjc.initialize_state(T=np.array([298.0]))
        results_yjc = self.model_yjc.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_yjc[-1, 2, 0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.01
        stress_FLAG = 317.06e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / stat.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_lowrate(self):
        """
        JC, constant g0 and Tm, low rate 1e-1/s, FLAG/python comparison within 5%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-7(micro-s)^-1 = 1e-1(s)^-1

        FLAG sim results
        eps_p = 0.28928, flow_stress =  284.89 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.28928]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_yjc.initialize(self.params, self.consts)
        self.model_yjc.initialize_state(T=np.array([298.0]))
        results_yjc = self.model_yjc.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_yjc[-1, 2, 0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.05
        stress_FLAG = 284.89e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / stat.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)





# from models_withLik import interpolate_experiment
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

import impala.superCal as impala

plt.rcParams.update({"figure.max_open_warning": 0})
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.lines import Line2D
from numpy.random import uniform
from scipy.interpolate import interp1d
from scipy.stats import gaussian_kde


# diagnostic plots to include:
#    - traces of theta, s2, but for hier which ones? All?
#   - tempering and MCMC counts
class Diagnostics(object):
    def __init__(self, setup, out):
        """ """
        self.setup = setup
        self.out = out
        return

    def pooled_trace_plots():
        pass


class PTW_Plotter(object):
    """PTW Prediction and Pairwise Plots"""

    pooled = None

    def __init__(self, setup, out):
        """ """
        self.setup = setup
        self.out = out
        return

    def ptw_prediction_plot_single(
        self,
        actual_x,
        actual_y,
        pred_y_lb,
        pred_y_ub,
        hier_y_lb=None,
        hier_y_ub=None,
        pnew_y_lb=None,
        pnew_y_ub=None,
        edot=None,
        temp=None,
        pdf=None,
        model_type=None,
    ):
        """PTW Prediction Single Plot"""
        ymaxs = [actual_y.max(), pred_y_ub.max()]
        if hier_y_ub is not None:
            ymaxs.append(hier_y_ub.max())
        if pnew_y_ub is not None:
            ymaxs.append(pnew_y_ub.max())
        ymax = max(ymaxs)
        xmax = actual_x.max()
        text_coords = (0.1 * xmax, 0.85 * ymax)

        fig = plt.figure(figsize=(4, 3))
        if pnew_y_lb is not None:
            plt.fill_between(
                actual_x,
                pnew_y_lb,
                pnew_y_ub,
                color="lightgrey",
                label=r"$\theta^*$",
            )
        if hier_y_lb is not None:
            plt.fill_between(
                actual_x,
                hier_y_lb,
                hier_y_ub,
                color="lightblue",
                label=r"$\theta_0$",
            )
        plt.fill_between(
            actual_x,
            pred_y_lb,
            pred_y_ub,
            color="lightgreen",
            label=r"$\theta_i$",
        )
        plt.scatter(actual_x, actual_y, color="blue", s=0.5, label="y")
        plt.ylim(0.0, 0.025)
        if edot is not None and temp is not None:
            plt.text(*text_coords, "edot: {:.3E}\n temp: {}".format(edot, temp))
        if pdf:
            pdf.savefig(fig)
        return

    def ptw_prediction_plots_hier(self, path, uu):
        """PTW Prediction Hierarchical Plots (no input)"""
        sel = uu  # np.arange(20000, self.setup.nmcmc, 10) # need to script this in
        pred_theta_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]
        pred_theta0_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]
        pred_thetap_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]

        theta_parent = impala.chol_sample_1per_constraints(
            self.out.theta0[sel, 0],
            self.out.Sigma0[sel, 0],
            self.setup.checkConstraints,
            self.setup.bounds_mat,
            self.setup.bounds.keys(),
            self.setup.bounds,
            self.setup.constants,
        )

        for i in range(self.setup.nexp):
            self.setup.models[i].pool = False
            pred_theta_raw[i] = self.setup.models[i].eval(
                impala.tran_unif(
                    self.out.theta[i][sel, 0, :, :].reshape([-1, self.setup.p]),
                    self.setup.bounds_mat,
                    self.setup.bounds.keys(),
                ),
                pool=False,
                nugget=True,
            )
            pred_theta0_raw[i] = self.setup.models[i].eval(
                impala.tran_unif(
                    self.out.theta0[sel, 0],
                    self.setup.bounds_mat,
                    self.setup.bounds.keys(),
                ),
                pool=True,
                nugget=True,
            )
            pred_thetap_raw[i] = self.setup.models[i].eval(
                impala.tran_unif(
                    theta_parent,
                    self.setup.bounds_mat,
                    self.setup.bounds.keys(),
                ),
                pool=True,
                nugget=True,
            )
            self.setup.models[i].pool = True
        real_strain = []
        real_stress = []
        pred_theta = []
        pred_theta0 = []
        pred_thetap = []
        pred_theta_quant_lb = []
        pred_theta_quant_ub = []
        pred_theta0_quant_lb = []
        pred_theta0_quant_ub = []
        pred_thetap_quant_lb = []
        pred_thetap_quant_ub = []
        edots = []
        temps = []
        model_type = []

        for i in range(self.setup.nexp):
            for j in range(self.setup.ntheta[i]):
                pred_theta.append(
                    pred_theta_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_theta_quant_lb.append(
                    np.quantile(pred_theta[-1], 0.005, 0)
                )
                pred_theta_quant_ub.append(
                    np.quantile(pred_theta[-1], 0.995, 0)
                )
                pred_theta0.append(
                    pred_theta0_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_theta0_quant_lb.append(
                    np.quantile(pred_theta0[-1], 0.005, 0)
                )
                pred_theta0_quant_ub.append(
                    np.quantile(pred_theta0[-1], 0.995, 0)
                )
                pred_thetap.append(
                    pred_thetap_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_thetap_quant_lb.append(
                    np.quantile(pred_thetap[-1], 0.005, 0)
                )
                pred_thetap_quant_ub.append(
                    np.quantile(pred_thetap[-1], 0.995, 0)
                )
                model_type.append(type(self.setup.models[i]).__name__)
                if model_type[-1] == "ModelMaterialStrength":
                    edots.append(self.setup.models[i].edots[j])
                    temps.append(self.setup.models[i].temps[j])
                    real_strain.append(
                        self.setup.models[i].meas_strain_histories[j]
                    )
                    real_stress.append(
                        self.setup.ys[i][self.setup.s2_ind[i] == j]
                    )
                else:
                    edots.append(None)
                    temps.append(None)
                    real_strain.append(
                        np.linspace(
                            0,
                            1,
                            len(self.setup.ys[i][self.setup.s2_ind[i] == j]),
                        )
                    )
                    real_stress.append(
                        self.setup.ys[i][self.setup.s2_ind[i] == j]
                    )

        rows = zip(
            real_strain,
            real_stress,
            pred_theta_quant_lb,
            pred_theta_quant_ub,
            pred_theta0_quant_lb,
            pred_theta0_quant_ub,
            pred_thetap_quant_lb,
            pred_thetap_quant_ub,
            edots,
            temps,
            model_type,
        )
        keys = [
            "actual_x",
            "actual_y",
            "pred_y_lb",
            "pred_y_ub",
            "hier_y_lb",
            "hier_y_ub",
            "pnew_y_lb",
            "pnew_y_ub",
            "edot",
            "temp",
            "model_type",
        ]
        plot_param_list = [dict(zip(keys, row)) for row in rows]

        pdf = PdfPages(path)
        for plot_params in plot_param_list:
            if plot_params["model_type"] == "ModelMaterialStrength":
                self.ptw_prediction_plot_single(**plot_params, pdf=pdf)
            else:
                if len(plot_params["actual_x"]) == 1:
                    pass
                else:
                    self.ptw_prediction_plot_single(**plot_params, pdf=pdf)
        pdf.close()
        return

    def ptw_prediction_plots_pool(self, path, uu):
        """PTW Prediction Hierarchical Plots (no input)"""
        sel = uu  # np.arange(20000, self.setup.nmcmc, 10) # need to script this in
        pred_theta_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]
        for i in range(self.setup.nexp):
            pred_theta_raw[i] = self.setup.models[i].eval(
                impala.tran_probit(
                    self.out.theta[
                        sel, 0
                    ],  # .repeat(self.setup.ns2[i], axis = 0),
                    self.setup.bounds_mat,
                    self.setup.bounds.keys(),
                ),
            )

        real_strain = []
        real_stress = []
        pred_theta = []
        pred_theta_quant_lb = []
        pred_theta_quant_ub = []
        edots = []
        temps = []
        model_type = []

        for i in range(self.setup.nexp):
            for j in range(self.setup.ns2[i]):
                pred_theta.append(
                    pred_theta_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_theta_quant_lb.append(
                    np.quantile(pred_theta[-1], 0.025, 0)
                )
                pred_theta_quant_ub.append(
                    np.quantile(pred_theta[-1], 0.975, 0)
                )
                model_type.append(type(self.setup.models[i]).__name__)
                if model_type[-1] == "ModelMaterialStrength":
                    edots.append(self.setup.models[i].edots[j])
                    temps.append(self.setup.models[i].temps[j])
                    real_strain.append(
                        self.setup.models[i].meas_strain_histories[j]
                    )
                    real_stress.append(
                        self.setup.ys[i][self.setup.s2_ind[i] == j]
                    )
                else:
                    edots.append(None)
                    temps.append(None)
                    real_strain.append(
                        np.linspace(
                            0,
                            1,
                            len(self.setup.ys[i][self.setup.s2_ind[i] == j]),
                        )
                    )
                    real_stress.append(
                        self.setup.ys[i][self.setup.s2_ind[i] == j]
                    )

        rows = zip(
            real_strain,
            real_stress,
            pred_theta_quant_lb,
            pred_theta_quant_ub,
            edots,
            temps,
            model_type,
        )
        keys = [
            "actual_x",
            "actual_y",
            "pred_y_lb",
            "pred_y_ub",
            "edot",
            "temp",
            "model_type",
        ]
        plot_param_list = [dict(zip(keys, row)) for row in rows]

        pdf = PdfPages(path)
        for plot_params in plot_param_list:
            if plot_params["model_type"] == "ModelMaterialStrength":
                self.ptw_prediction_plot_single(**plot_params, pdf=pdf)
            else:
                if len(plot_params["actual_x"]) == 1:
                    pass
                else:
                    self.ptw_prediction_plot_single(**plot_params, pdf=pdf)
        pdf.close()
        return

    def ptw_prediction_plots_cluster(self, path):
        sel = np.arange(20000, self.setup.nmcmc, 10)  # need to script this in
        pred_theta_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]
        pred_theta0_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]
        pred_thetap_raw = [
            np.empty([sel.shape[0], self.setup.ys[i].shape[0]])
            for i in range(self.setup.nexp)
        ]

        thetas = self.out.theta[sel, 0]
        deltas = [self.out.delta[i][sel] for i in range(self.setup.nexp)]
        nclustmax = (
            max(self.out.delta[i].max() for i in range(self.setup.nexp)) + 1
        )
        dcounts = np.zeros((sel.shape[0], nclustmax))
        for it, s in enumerate(sel):
            for i in range(self.setup.nexp):
                dcounts[it] += np.bincount(
                    self.out.delta[i][s, 0], minlength=nclustmax
                )
        etas = self.out.eta[sel, 0]
        nclust = (dcounts > 0).sum(axis=1)
        prob = dcounts + (dcounts == 0) * (
            etas / (nclustmax - nclust + 1e-9)
        ).reshape(-1, 1)
        prob[:] /= prob.sum(axis=1).reshape(-1, 1)
        cumprob = np.cumsum(prob, axis=1)
        unif = uniform(size=(sel.shape[0], 1))
        dnew = (unif > cumprob).sum(axis=1)
        theta_parent = thetas[np.arange(sel.shape[0]), dnew]
        for i in range(self.setup.nexp):
            for j in range(sel.shape[0]):
                pred_theta_raw[i][j] = self.setup.models[i].eval(
                    impala.tran_unif(
                        self.out.theta_hist[i][sel[j], 0],
                        self.setup.bounds_mat,
                        self.setup.bounds.keys(),
                    ),
                )
                pred_theta0_raw[i][j] = self.setup.models[i].eval(
                    impala.tran_unif(
                        np.repeat(
                            self.out.theta0[sel[j], 0].reshape(1, -1),
                            self.setup.ns2[i],
                            axis=0,
                        ),
                        self.setup.bounds_mat,
                        self.setup.bounds.keys(),
                    ),
                )
                pred_thetap_raw[i][j] = self.setup.models[i].eval(
                    impala.tran_unif(
                        np.repeat(
                            theta_parent[j].reshape(1, -1),
                            self.setup.ns2[i],
                            axis=0,
                        ),
                        self.setup.bounds_mat,
                        self.setup.bounds.keys(),
                    ),
                )

        real_strain = []
        real_stress = []
        pred_theta = []
        pred_theta0 = []
        pred_thetap = []
        pred_theta_quant_lb = []
        pred_theta_quant_ub = []
        pred_theta0_quant_lb = []
        pred_theta0_quant_ub = []
        pred_thetap_quant_lb = []
        pred_thetap_quant_ub = []
        edots = []
        temps = []

        for i in range(self.setup.nexp):
            for j in range(self.setup.ntheta[i]):
                pred_theta.append(
                    pred_theta_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_theta_quant_lb.append(
                    np.quantile(pred_theta[-1], 0.025, 0)
                )
                pred_theta_quant_ub.append(
                    np.quantile(pred_theta[-1], 0.975, 0)
                )
                pred_theta0.append(
                    pred_theta0_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_theta0_quant_lb.append(
                    np.quantile(pred_theta0[-1], 0.025, 0)
                )
                pred_theta0_quant_ub.append(
                    np.quantile(pred_theta0[-1], 0.975, 0)
                )
                pred_thetap.append(
                    pred_thetap_raw[i].T[self.setup.s2_ind[i] == j].T
                )
                pred_thetap_quant_lb.append(
                    np.quantile(pred_thetap[-1], 0.025, 0)
                )
                pred_thetap_quant_ub.append(
                    np.quantile(pred_thetap[-1], 0.975, 0)
                )
                edots.append(self.setup.models[i].edots[j])
                temps.append(self.setup.models[i].temps[j])
                real_strain.append(
                    self.setup.models[i].meas_strain_histories[j]
                )
                real_stress.append(self.setup.ys[i][self.setup.s2_ind[i] == j])

        rows = zip(
            real_strain,
            real_stress,
            pred_theta_quant_lb,
            pred_theta_quant_ub,
            pred_theta0_quant_lb,
            pred_theta0_quant_ub,
            pred_thetap_quant_lb,
            pred_thetap_quant_ub,
            edots,
            temps,
        )
        keys = [
            "actual_x",
            "actual_y",
            "pred_y_lb",
            "pred_y_ub",
            "hier_y_lb",
            "hier_y_ub",
            "pnew_y_lb",
            "pnew_y_ub",
            "edot",
            "temp",
        ]
        plot_param_list = [dict(zip(keys, row)) for row in rows]

        pdf = PdfPages(path)
        for plot_params in plot_param_list:
            self.ptw_prediction_plot_single(**plot_params, pdf=pdf)
        pdf.close()
        return

    def ptw_prediction_plots(self, path):
        """PTW Prediction Plots against model"""
        if type(self.out) is impala.OutCalibPool:
            return self.ptw_prediction_plots_pool(path)
        elif type(self.out) is impala.OutCalibHier:
            return self.ptw_prediction_plots_hier(path)
        elif type(self.out) is impala.OutCalibClust:
            return self.ptw_prediction_plots_cluster(path)
        else:
            raise ValueError("Improper out type")
        return

    @staticmethod
    def kde_contour(x1, x2, percentile):
        density = gaussian_kde([x1, x2], bw_method="silverman")
        X, Y = np.mgrid[min(x1) : max(x1) : 100j, min(x2) : max(x2) : 100j]
        positions = np.vstack([X.ravel(), Y.ravel()])
        Z = density(positions)
        Z /= Z.sum()
        t = np.linspace(0, Z.max(), 1000)
        integral = ((Z >= t[:, None, None]) * Z).sum(axis=(1, 2))
        f = interp1d(integral, t)
        t_contours = f(np.array([percentile]))
        return {"X": X, "Y": Y, "Z": Z.reshape([100, 100]), "conts": t_contours}

    def pairwise_theta_plot_hier(self, path, uu, highlight=None):
        """Pairwise Theta scatterplot"""
        if highlight is None:
            highlight = [
                range(self.setup.ntheta[k]) for k in range(self.setup.nexp)
            ]
        sel = uu  # np.arange(20000, self.setup.nmcmc, 10)
        theta_parent = impala.chol_sample_1per_constraints(
            self.out.theta0[sel, 0],
            self.out.Sigma0[sel, 0],
            self.setup.checkConstraints,
            self.setup.bounds_mat,
            self.setup.bounds.keys(),
            self.setup.bounds,
            self.setup.constants,
        )
        theta_names = list(self.setup.bounds.keys())
        theta0_unst = impala.unnormalize(
            self.out.theta0[sel, 0, :], self.setup.bounds_mat
        )
        theta_parent_unst = impala.unnormalize(
            theta_parent, self.setup.bounds_mat
        )
        theta_unst = [
            self.out.theta[k][sel, 0, :, :] for k in range(self.setup.nexp)
        ]
        for k in range(self.setup.nexp):
            for s in range(self.setup.ntheta[k]):
                theta_unst[k][:, s, :] = impala.unnormalize(
                    self.out.theta[k][sel, 0, s, :], self.setup.bounds_mat
                )
        plt.figure(figsize=(15, 15))
        for i in range(self.setup.p):
            for j in range(self.setup.p):
                if i == j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    for k in range(self.setup.nexp):
                        # for s in range(self.setup.ntheta[k]):
                        if highlight[k] is not None:
                            for s in highlight[k]:
                                sns.kdeplot(
                                    theta_unst[k][:, s, i], color="lightgreen"
                                )
                    sns.kdeplot(theta0_unst[:, i], color="blue")
                    sns.kdeplot(theta_parent_unst[:, i], color="grey")
                    # plt.xlim(0,1)

                    plt.xlim(
                        self.setup.bounds_mat[i, 0], self.setup.bounds_mat[i, 1]
                    )

                    ax = plt.gca()
                    ax.axes.yaxis.set_visible(False)
                    plt.xlabel(theta_names[i])
                    ax.tick_params(axis="x", which="major", labelsize=8)
                    plt.setp(
                        ax.get_xticklabels(),
                        rotation=30,
                        horizontalalignment="right",
                    )
                elif i < j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    for k in range(self.setup.nexp):
                        # for s in range(self.setup.ntheta[k]):
                        if highlight[k] is not None:
                            for s in highlight[k]:
                                contour = self.kde_contour(
                                    theta_unst[k][:, s, j],
                                    theta_unst[k][:, s, i],
                                    0.9,
                                )
                                plt.contour(
                                    contour["X"],
                                    contour["Y"],
                                    contour["Z"],
                                    contour["conts"],
                                    colors="lightgreen",
                                )
                    contour = self.kde_contour(
                        theta0_unst[:, j], theta0_unst[:, i], 0.9
                    )
                    plt.contour(
                        contour["X"],
                        contour["Y"],
                        contour["Z"],
                        contour["conts"],
                        colors="blue",
                    )

                    contour = self.kde_contour(
                        theta_parent_unst[:, j], theta_parent_unst[:, i], 0.9
                    )
                    plt.contour(
                        contour["X"],
                        contour["Y"],
                        contour["Z"],
                        contour["conts"],
                        colors="grey",
                    )
                    # plt.xlim(0,1)
                    # plt.ylim(0,1)

                    plt.xlim(
                        self.setup.bounds_mat[j, 0], self.setup.bounds_mat[j, 1]
                    )
                    plt.ylim(
                        self.setup.bounds_mat[i, 0], self.setup.bounds_mat[i, 1]
                    )

                    ax = plt.gca()
                    ax.axes.xaxis.set_visible(False)
                    ax.axes.yaxis.set_visible(False)
                else:
                    pass
        plt.subplots_adjust(wspace=0.05, hspace=0.05)
        plt.subplot2grid((self.setup.p, self.setup.p), (2, 0))
        colors = ["lightgreen", "blue", "grey"]
        lines = [Line2D([0], [0], color=c, linewidth=2) for c in colors]
        labels = [r"$\theta_i$", r"$\theta_0$", r"$\theta^*$"]
        plt.legend(lines, labels)
        plt.axis("off")
        if path:
            plt.savefig(path, bbox_inches="tight")
        else:
            plt.show()
        return

    def pairwise_theta_plot_pool(self, path, uu):
        sel = uu  # np.arange(20000, self.setup.nmcmc, 10)
        plt.figure(figsize=(15, 15))
        for i in range(self.setup.p):
            for j in range(self.setup.p):
                if i == j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    sns.distplot(
                        impala.invprobit(self.out.theta[sel, 0, i]),
                        hist=False,
                        kde=True,
                        color="blue",
                    )
                    plt.xlim(0, 1)
                    ax = plt.gca()
                    ax.axes.yaxis.set_visible(False)
                    ax.tick_params(axis="x", which="major", labelsize=8)
                    plt.setp(
                        ax.get_xticklabels(),
                        rotation=30,
                        horizontalalignment="right",
                    )
                elif i < j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    contour = self.kde_contour(
                        impala.invprobit(self.out.theta[sel, 0, j]),
                        impala.invprobit(self.out.theta[sel, 0, i]),
                        0.9,
                    )
                    plt.contour(
                        contour["X"],
                        contour["Y"],
                        contour["Z"],
                        contour["conts"],
                        colors="blue",
                    )
                    plt.xlim(0, 1)
                    plt.ylim(0, 1)
                    ax = plt.gca()
                    ax.axes.xaxis.set_visible(False)
                    ax.axes.yaxis.set_visible(False)
                else:
                    pass
        plt.subplots_adjust(wspace=0.05, hspace=0.05)
        plt.axis("off")
        if path:
            plt.savefig(path, bbox_inches="tight")
        else:
            plt.show()
        return

    def pairwise_theta_plot_cluster(self, path=None):
        sel = np.arange(20000, self.setup.nmcmc, 10)
        thetas = self.out.theta[sel, 0]
        deltas = [self.out.delta[i][sel] for i in range(self.setup.nexp)]
        nclustmax = (
            max(self.out.delta[i].max() for i in range(self.setup.nexp)) + 1
        )
        dcounts = np.zeros((sel.shape[0], nclustmax))
        for it, s in enumerate(sel):
            for i in range(self.setup.nexp):
                dcounts[it] += np.bincount(
                    self.out.delta[i][s, 0], minlength=nclustmax
                )
        etas = self.out.eta[sel, 0]
        nclust = (dcounts > 0).sum(axis=1)
        prob = dcounts + (dcounts == 0) * (
            etas / (nclustmax - nclust + 1e-9)
        ).reshape(-1, 1)
        prob /= prob.sum(axis=1).reshape(-1, 1)
        cumprob = np.cumsum(prob, axis=1)
        unif = uniform(size=(sel.shape[0], 1))
        dnew = (unif > cumprob).sum(axis=1)
        theta_parent = thetas[np.arange(sel.shape[0]), dnew]
        plt.figure(figsize=(15, 15))
        for i in range(self.setup.p):
            for j in range(self.setup.p):
                if i == j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    for k in range(self.setup.nexp):
                        for s in range(self.setup.ns2[k]):
                            sns.distplot(
                                impala.invprobit(
                                    self.out.theta_hist[k][sel, 0, s, i]
                                ),
                                hist=False,
                                kde=True,
                                color="lightgreen",
                            )
                    sns.distplot(
                        impala.invprobit(self.out.theta0[sel, 0, i]),
                        hist=False,
                        kde=True,
                        color="blue",
                    )
                    sns.distplot(
                        impala.invprobit(theta_parent[:, i]),
                        hist=False,
                        kde=True,
                        color="grey",
                    )
                    plt.xlim(0, 1)
                    ax = plt.gca()
                    ax.axes.yaxis.set_visible(False)
                    ax.tick_params(axis="x", which="major", labelsize=8)
                    plt.setp(
                        ax.get_xticklabels(),
                        rotation=30,
                        horizontalalignment="right",
                    )
                elif i < j:
                    plt.subplot2grid((self.setup.p, self.setup.p), (i, j))
                    for k in range(self.setup.nexp):
                        for s in range(self.setup.ns2[k]):
                            contour = self.kde_contour(
                                impala.invprobit(
                                    self.out.theta_hist[k][sel, 0, s, j]
                                ),
                                impala.invprobit(
                                    self.out.theta_hist[k][sel, 0, s, i]
                                ),
                                0.9,
                            )
                            plt.contour(
                                contour["X"],
                                contour["Y"],
                                contour["Z"],
                                contour["conts"],
                                colors="lightgreen",
                            )
                    contour = self.kde_contour(
                        impala.invprobit(self.out.theta0[sel, 0, j]),
                        impala.invprobit(self.out.theta0[sel, 0, i]),
                        0.9,
                    )
                    plt.contour(
                        contour["X"],
                        contour["Y"],
                        contour["Z"],
                        contour["conts"],
                        colors="blue",
                    )

                    contour = self.kde_contour(
                        impala.invprobit(theta_parent[:, j]),
                        impala.invprobit(theta_parent[:, i]),
                        0.9,
                    )
                    plt.contour(
                        contour["X"],
                        contour["Y"],
                        contour["Z"],
                        contour["conts"],
                        colors="grey",
                    )
                    plt.xlim(0, 1)
                    plt.ylim(0, 1)
                    ax = plt.gca()
                    ax.axes.xaxis.set_visible(False)
                    ax.axes.yaxis.set_visible(False)
                else:
                    pass
        plt.subplots_adjust(wspace=0.05, hspace=0.05)
        plt.subplot2grid((self.setup.p, self.setup.p), (2, 0))
        colors = ["lightgreen", "blue", "grey"]
        lines = [Line2D([0], [0], color=c, linewidth=2) for c in colors]
        labels = [r"$\theta_i$", r"$\theta_0$", r"$\theta^*$"]
        plt.legend(lines, labels)
        plt.axis("off")
        try:
            plt.subplot2grid((self.setup.p, self.setup.p), (4, 0))
            sns.distplot(nclust, kde=True, color="blue")
            plt.xlim(0, nclustmax)
        except IndexError:
            pass
        if path:
            plt.savefig(path, bbox_inches="tight")
        else:
            plt.show()
        pass

    def pairwise_theta_plot(self, path=None):
        if type(self.out) is impala.OutCalibPool:
            return self.pairwise_theta_plot_pool(path)
        elif type(self.out) is impala.OutCalibHier:
            return self.pairwise_theta_plot_hier(path)
        elif type(self.out) is impala.OutCalibClust:
            return self.pairwise_theta_plot_cluster(path)
        else:
            raise ValueError("Improper out type")
        return

    @staticmethod
    def cluster_matrix(delta_list, ns2, nclustmax, nburn=20000, nthin=10):
        # subset delta to post burn-in
        delta_relist = [d[nburn::nthin] for d in delta_list]
        # Declare constants
        nsamp = delta_relist[0].shape[0]
        nexp = len(delta_relist)
        # create a combined delta array (for all experiments/vectorized experiments)
        # Boolean array, so (True iff member of cluster)
        breaks = np.hstack((0, np.cumsum(ns2)))
        bounds = [
            (breaks[i], breaks[i + 1]) for i in range(breaks.shape[0] - 1)
        ]
        delta_mat = np.empty(
            (delta_relist[0].shape[0], breaks[-1], nclustmax), dtype=bool
        )
        # Fill the combined delta array
        for i in range(breaks.shape[0] - 1):
            delta_mat[:, bounds[i][0] : bounds[i][1]] = delta_relist[i][
                :, 0, :, None
            ] == np.arange(nclustmax)
        # X = matrix of incidence (nsamp x nclust) ->  XX^t matrix of coincidence (nsamp x nsamp)
        # Average over all iterations -> matrix of average coincidence / shared cluster membership
        out = np.einsum("icp,iqp->icq", delta_mat, delta_mat).mean(axis=0)
        return out, breaks

    def cluster_matrix_plot(self, path=None, **kwargs):
        cmat, breaks = self.cluster_matrix(
            self.out.delta, self.setup.ns2, self.out.nclustmax, **kwargs
        )
        plt.matshow(cmat)
        if breaks.shape[0] > 1:
            for breakpoint in breaks[1:-1] - 0.5:
                plt.axhline(breakpoint, color="red", linestyle="--")
                plt.axvline(breakpoint, color="green", linestyle="--")
        plt.legend()
        if path:
            plt.savefig(path, bbox_inches="tight")
        else:
            plt.show()
        return

    pass


# EOF



import unittest

import numpy as np

from impala.physics import physical_models_vec as pmh


# Constant yield stress model
class TestPTWYieldStress_Constg0Tm(unittest.TestCase):
    def setUp(self):
        """
        Parameters are for OFHC copper.
        """
        self.params = {
            # PTW
            "theta": np.array([0.025]),
            "p": np.array([2.0]),
            "s0": np.array([0.0085]),
            "sInf": np.array([0.00055]),
            "kappa": np.array([0.11]),
            "lgamma": np.array([np.log(1e-5)]),
            "y0": np.array([0.0001]),
            "yInf": np.array([0.00009999]),
            "y1": np.array([0.094]),
            "y2": np.array([0.575]),
        }
        self.consts = {
            # PTW
            "beta": 0.25,
            "matomic": 63.546,
            "chi": 1.0,
            # Constant Spec. Heat
            "Cv0": 0.383e-5,
            # Constant Density
            "rho0": 8.9375,
            # Constant Melt Temp.
            "Tmelt0": 1625.0,
            # # Constant Shear Mod.
            "G0": 0.4578,
            # # Simple Shear Mod.
            # 'G0'     : 0.50889, # Cold shear
            # 'alpha'  : 0.21
            # # SG Shear Mod.
            # 'G0'     : 0.4578, # MBar, 300K Shear mod.
            # 'sgB'    : 3.8e-4, # K^-1
        }
        self.model_ptw_cg0Tm = pmh.MaterialModel(
            flow_stress_model=pmh.PTW_Yield_Stress,
        )

    def test_isothermal_lowrate(self):
        """
        PTW, const. g0, Tm. Rates less than 1e-6/us (~1/s), should be isothermal T=const
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.array([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertEqual(results_ptw_cg0Tm[0][3], results_ptw_cg0Tm[-1][3])

    def test_adiabatic_highrate(self):
        """
        PTW, const. g0, Tm. Rates greater than 1e-6/us (~1/s), temp. changes adiabatically
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.array([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertNotEqual(results_ptw_cg0Tm[0][3], results_ptw_cg0Tm[-1][3])

    def test_stress_highrate(self):
        """
        PTW, constant g0 and Tm, high rate 1e6/s, FLAG/python comparison within 1%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e0(micro-s)^-1 = 1e6(s)^-1

        FLAG sim results
        eps_p = 0.244376, flow_stress =  521.4 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.

        NOTE: Temperature affects from volumetric heating may be offsetting the
        affect of holding the shear modulus constant, thus producing a tight match
        with the Python result, although for likely non-physical reasons.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.244376]), edot=np.array([1e0]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.array([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_cg0Tm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.01
        stress_FLAG = 521.416e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midhighrate(self):
        """
        PTW, constant g0 and Tm, mid-high rate 1e4/s, FLAG/python comparison within 9%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-2(micro-s)^-1 = 1e4(s)^-1

        FLAG sim results
        eps_p = 0.382749, flow_stress =  458.198 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.382749]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.array([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_cg0Tm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.09
        stress_FLAG = 458.198e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midlowrate(self):
        """
        PTW, constant g0 and Tm, low-mid rate 1e2/s, FLAG/python comparison within 11%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-4(micro-s)^-1 = 1e2(s)^-1

        FLAG sim results
        eps_p = 0.388339, flow_stress =  418.512 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.388339]), edot=np.array([1e-4]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.array([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_cg0Tm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.11
        stress_FLAG = 418.512e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_lowrate(self):
        """
        PTW, constant g0 and Tm, low rate 1e0/s, FLAG/python comparison within 9%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-6(micro-s)^-1 = 1e0(s)^-1

        FLAG sim results
        eps_p = 0.371143, flow_stress =  385.751 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.371143]), edot=np.array([1e-6]), nhist=1000
        )
        self.model_ptw_cg0Tm.initialize(self.params, self.consts)
        self.model_ptw_cg0Tm.initialize_state(T=np.asarray([298.0]))
        results_ptw_cg0Tm = self.model_ptw_cg0Tm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_cg0Tm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.09
        stress_FLAG = 385.751e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)


class TestPTWYieldStress_SimpShearConstTm(unittest.TestCase):
    def setUp(self):
        """
        Parameters are for OFHC copper.
        """
        self.params = {
            # PTW
            "theta": np.array([0.025]),
            "p": np.array([2.0]),
            "s0": np.array([0.0085]),
            "sInf": np.array([0.00055]),
            "kappa": np.array([0.11]),
            "lgamma": np.array([np.log(1e-5)]),
            "y0": np.array([0.0001]),
            "yInf": np.array([0.00009999]),
            "y1": np.array([0.094]),
            "y2": np.array([0.575]),
        }
        self.consts = {
            # PTW
            "beta": 0.25,
            "matomic": 63.546,
            "chi": 1.0,
            # Constant Spec. Heat
            "Cv0": 0.383e-5,
            # Constant Density
            "rho0": 8.9375,
            # Constant Melt Temp.
            "Tmelt0": 1625.0,
            # # Constant Shear Mod.
            # 'G0'     : 0.4578,
            # Simple Shear Mod.
            "G0": 0.50889,  # Cold shear
            "alpha": 0.21,
            # # SG Shear Mod.
            # 'G0'     : 0.4578, # MBar, 300K Shear mod.
            # 'sgB'    : 3.8e-4, # K^-1
        }
        self.model_ptw_ss_cTm = pmh.MaterialModel(
            flow_stress_model=pmh.PTW_Yield_Stress,
            shear_modulus_model=pmh.Simple_Shear_Modulus,
        )

    def test_isothermal_lowrate(self):
        """
        PTW, const. Tm, PW shear. Rates less than 1e-6/us (~1/s), should be isothermal T=const
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertEqual(results_ptw_ss_cTm[0][3], results_ptw_ss_cTm[-1][3])

    def test_adiabatic_highrate(self):
        """
        PTW, const. Tm, PW shear. Rates greater than 1e-6/us (~1/s), temp. changes adiabatically
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertNotEqual(results_ptw_ss_cTm[0][3], results_ptw_ss_cTm[-1][3])

    def test_stress_highrate(self):
        """
        PTW, constant Tm, simple shear mod, high rate 1e6/s, temp.=620K, FLAG/python comparison within 6%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e0(micro-s)^-1 = 1e6(s)^-1

        FLAG sim results
        eps_p = 0.244463, flow_stress =  485.301 MPa, temp ~ 620 K

        Results are expected to be different at this high rate of strain due
        to the fact that there are temperature effects from volumetric strain
        that significantly affect the shear modulus and flow stress. A trick
        would be to initialize the model closer to the temperatures expressed
        by FLAG at maximum strain. This could get the python result closer to
        the FLAG result
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.244463]), edot=np.array([1e0]), nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([620.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_ss_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.06
        stress_FLAG = 485.301e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midhighrate(self):
        """
        PTW, constant Tm, simple shear mod, mid-high rate 1e4/s, FLAG/python comparison within 2%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-2(micro-s)^-1 = 1e4(s)^-1

        FLAG sim results
        eps_p = 0.383542, flow_stress =  439.790 MPa, temp ~ 341.5 K

        See note above from the high rate test. Temperature effects may still
        create significant differences in computed stress.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.383542]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_ss_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.02
        stress_FLAG = 439.790e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midlowrate(self):
        """
        PTW, constant Tm, simple shear mod, low-mid rate 1e2/s, FLAG/python comparison within 1%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-4(micro-s)^-1 = 1e2(s)^-1

        FLAG sim results
        eps_p = 0.388302, flow_stress =  401.108 MPa, temp ~ 328.1 K

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.388302]), edot=1e-4, nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_ss_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.01
        stress_FLAG = 401.108e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_lowrate(self):
        """
        PTW, constant Tm, simple shear mod, low rate 1e0/s, FLAG/python comparison within 2%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-6(micro-s)^-1 = 1e0(s)^-1

        FLAG sim results
        eps_p = 0.385416, flow_stress =  375.227 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.385416]), edot=np.array([1e-6]), nhist=1000
        )
        self.model_ptw_ss_cTm.initialize(self.params, self.consts)
        self.model_ptw_ss_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_ss_cTm = self.model_ptw_ss_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_ss_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.02
        stress_FLAG = 375.227e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)


class TestPTWYieldStress_SteinShearConstTm(unittest.TestCase):
    def setUp(self):
        """
        Parameters are for OFHC copper.
        """
        self.params = {
            # PTW
            "theta": np.array([0.025]),
            "p": np.array([2.0]),
            "s0": np.array([0.0085]),
            "sInf": np.array([0.00055]),
            "kappa": np.array([0.11]),
            "lgamma": np.array([np.log(1e-5)]),
            "y0": np.array([0.0001]),
            "yInf": np.array([0.00009999]),
            "y1": np.array([0.094]),
            "y2": np.array([0.575]),
        }
        self.consts = {
            # PTW
            "beta": 0.25,
            "matomic": 63.546,
            "chi": 1.0,
            # Constant Spec. Heat
            "Cv0": 0.383e-5,
            # Constant Density
            "rho0": 8.9375,
            # Constant Melt Temp.
            "Tmelt0": 1625.0,
            # # Constant Shear Mod.
            # 'G0'     : 0.4578,
            # # Simple Shear Mod.
            # 'G0'     : 0.50889, # Cold shear
            # 'alpha'  : 0.21
            # SG Shear Mod.
            "G0": 0.4578,  # MBar, 300K Shear mod.
            "sgB": 3.8e-4,  # K^-1
        }
        self.model_ptw_sg_cTm = pmh.MaterialModel(
            flow_stress_model=pmh.PTW_Yield_Stress,
            shear_modulus_model=pmh.Stein_Shear_Modulus,
        )

    def test_isothermal_lowrate(self):
        """
        PTW, const. Tm, SG shear. Rates less than 1e-6/us (~1/s), should be isothermal T=const
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-7]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertEqual(results_ptw_sg_cTm[0][3], results_ptw_sg_cTm[-1][3])

    def test_adiabatic_highrate(self):
        """
        PTW, const. Tm, SG shear. Rates greater than 1e-6/us (~1/s), temp. changes adiabatically
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([1.0]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        self.assertNotEqual(results_ptw_sg_cTm[0][3], results_ptw_sg_cTm[-1][3])

    def test_stress_highrate(self):
        """
        PTW, constant Tm, Steinberg-Guinan shear mod, high rate 1e6/s, temp.=617K, FLAG/python comparison within 4%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e0(micro-s)^-1 = 1e6(s)^-1

        FLAG sim results
        eps_p = 0.244632, flow_stress =  427.011 MPa, temp ~ 617 K

        Results are expected to be different at this high rate of strain due
        to the fact that there are temperature effects from volumetric strain
        that significantly affect the shear modulus and flow stress. A trick
        would be to initialize the model closer to the temperatures expressed
        by FLAG at maximum strain. This could get the python result closer to
        the FLAG result
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.244632]), edot=np.array([1e0]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([617.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_sg_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.04
        stress_FLAG = 427.011e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midhighrate(self):
        """
        PTW, constant Tm, Steinberg-Guinan shear mod, mid-high rate 1e4/s, FLAG/python comparison within 2%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-2(micro-s)^-1 = 1e4(s)^-1

        FLAG sim results
        eps_p = 0.384834, flow_stress =  410.483 MPa, temp ~ 339 K

        See note above from the high rate test. Temperature effects may still
        create significant differences in computed stress.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.384834]), edot=np.array([1e-2]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_sg_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.02
        stress_FLAG = 410.483e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_midlowrate(self):
        """
        PTW, constant Tm, Steinberg-Guinan shear mod, low-mid rate 1e2/s, FLAG/python comparison within 1%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-4(micro-s)^-1 = 1e2(s)^-1

        FLAG sim results
        eps_p = 0.388306, flow_stress =  373.629 MPa, temp ~ 326 K

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.388306]), edot=np.array([1e-4]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_sg_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.01
        stress_FLAG = 373.629e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)

    def test_stress_lowrate(self):
        """
        PTW, constant Tm, Steinberg-Guinan shear mod, low rate 1e0/s, FLAG/python comparison within 2%.

        Compare flow stress with 3D single cell uniaxial compression simulation
        performed with FLAG. 2D and 1D produced multiaxial stress states as
        these dimensional spaces assume plane strain on the non-simulated
        dimension(s). Constant shear stress is the only node available with
        Johnson-Cook flow stress model in FLAG.

        High rate = 1e-6(micro-s)^-1 = 1e0(s)^-1

        FLAG sim results
        eps_p = 0.388966, flow_stress =  351.103 MPa

        Results will not be exact due to numeric differences. Therefore,
        perform assertion that Python result is within certain percentage of
        FLAG result at an advanced level of plastic strain.

        This makes it difficult to directly compare the initial yield stress at
        zero plastic strain.
        """
        shist = pmh.generate_strain_history_new(
            emax=np.array([0.388966]), edot=np.array([1e-6]), nhist=1000
        )
        self.model_ptw_sg_cTm.initialize(self.params, self.consts)
        self.model_ptw_sg_cTm.initialize_state(T=np.array([298.0]))
        results_ptw_sg_cTm = self.model_ptw_sg_cTm.compute_state_history(shist)
        # result format
        # [time, strain, stress, temp, shear_mod, density]
        # Assess stress at the emax
        stress_emax = results_ptw_sg_cTm[-1, 2][0]
        # Set maximum relative tolerance between FLAG and Python
        stress_tol = 0.02
        stress_FLAG = 351.103e-5  # Units MBar
        rel_diff = abs(stress_emax - stress_FLAG) / np.mean([
            stress_emax,
            stress_FLAG,
        ])

        self.assertTrue(rel_diff < stress_tol)


if __name__ == "__main__":
    unittest.main()



import multiprocessing as mp
import time
from collections import namedtuple
from math import sqrt

import numpy as np
import scipy
from numpy.linalg import slogdet
from numpy.random import beta, choice, gamma, uniform
from scipy.stats import invwishart

from impala.superCal.impala_noProbit_emu import *

# import pbar
# np.seterr(under='ignore')

# no probit tranform for hierarchical and DP versions


##############################################################################################################################################################################
def bincount2D_vectorized(a, max_count):
    """
    Applies np.bincount across a 2d array (row-wise).

    Adapted From: https://stackoverflow.com/questions/46256279/bin-elements-per-row-vectorized-2d-bincount-for-numpy
    """
    a_offs = a + np.arange(a.shape[0])[:, None] * max_count
    return np.bincount(
        a_offs.ravel(), minlength=a.shape[0] * max_count
    ).reshape(-1, max_count)


def sample_delta_per_temperature(
    curr_delta_t, njs_t, log_posts_t, inv_temp_t, eta_t
):
    """Given log-posteriors, current delta --
    iterates through delta assignments, reassigning probabilistically."""
    djs = (njs_t == 0) * eta_t / (sum(njs_t == 0) + 1e-9)
    temp = np.empty(log_posts_t.shape[0])
    unis = uniform(size=curr_delta_t.shape[0])
    for s in range(curr_delta_t.shape[0]):
        njs_t[curr_delta_t[s]] -= 1
        temp[:] = -np.inf
        temp[(njs_t + djs) > 0] = (  # log-posterior under each cluster
            inv_temp_t
            * (
                np.log((njs_t + djs)[njs_t + djs > 0])
                + log_posts_t.T[s, njs_t + djs > 0]
            )
        )
        temp[:] -= temp.max()  #
        # temp[:] = np.exp(temp) / np.exp(temp).sum()
        temp[:] = np.cumsum(np.exp(temp))
        temp[:] /= temp[
            -1
        ]  # normalized cumulative sum of probability of cluster choice
        curr_delta_t[s] = (unis[s] >= temp).sum()
        # curr_delta_t[s] = choice(log_posts_t.shape[0], p = temp)
        njs_t[curr_delta_t[s]] += 1
        djs[njs_t > 0] = 0.0
    return curr_delta_t


def sample_delta_per_temperature_wrapper(args):
    return sample_delta_per_temperature(*args)


def sample_delta(curr_delta, njs, log_posts, inv_temps, eta, pool=None):
    args = zip(curr_delta, njs, log_posts, inv_temps, eta)
    if pool is None:
        delta = np.array(list(map(sample_delta_per_temperature_wrapper, args)))
    else:
        delta = np.array(
            list(pool.map(sample_delta_per_temperature_wrapper, args))
        )
    return delta


def one_step_in_cluster_covariance_update(Sj, mSj, nSj, n, mus, covs):
    """
    Online update of covariance matrix -- as stepping through cluster assignments
    S, m, n -- modified in place

    Sj: view of covariance matrix
    Adapted from: https://tinyurl.com/onlinecovariance
    https://www.mathworks.com/help/matlab/import_export/using-mapreduce-to-compute-covariance-and-related-quantities.html
    Sj   is [ntemps x p x p]
    nSj  is [ntemps]
    mSj  is [ntemps]
    mus  is [ntemps x p]
    covs is [ntemps x p x p]
    """
    nC = nSj + n  # ntemps
    mC = (nSj.reshape(-1, 1) * mSj + n * mus) / nC.reshape(-1, 1)  # ntemps x p
    Sj[:] = (
        1
        / nC.reshape(-1, 1, 1)
        * (
            +nSj.reshape(-1, 1, 1) * Sj
            + n * covs
            + np.einsum("t,tp,tq->tpq", nSj, mSj - mC, mSj - mC)
            + n * np.einsum("tp,tq->tpq", mus - mC, mus - mC)
        )
    )
    mSj[:] = mC
    nSj[:] = nC
    return Sj, mSj, nSj


def cluster_covariance_update_old(
    S, mS, nS, n, delta, covs, mus, nexp, nclustmax, temps
):
    """
    S, m, n -- modified in place
    Goal is cluster-level covariance matrix
    input:
        means [i]x[ntemps, ns2[i], p],
        covs  [i]x[ntemps, ns2[i], p, p],
        delta [i]x[ntemps, ns2[i]]
    output:
        S     [ntemps, nclustmax, p, p]
    """
    S[:] = np.eye(S.shape[-1]) * 1e-9
    mS[:] = 0.0
    nS[:] = 0.0
    for i in range(nexp):
        for j in range(delta[i].shape[1]):
            (
                S[temps, delta[i].T[j]],
                mS[temps, delta[i].T[j]],
                nS[temps, delta[i].T[j]],
            ) = one_step_in_cluster_covariance_update(
                S[temps, delta[i].T[j]],
                mS[temps, delta[i].T[j]],
                nS[temps, delta[i].T[j]],
                n,
                mus[i][temps, j],
                covs[i][temps, j],
            )
    return


# West 1992, dirichletprocess R package
def sample_eta(curr_eta, nclust, ndat, prior_shape, prior_rate):
    g = beta(curr_eta + 1, ndat)
    aa = prior_shape + nclust
    bb = prior_rate - np.log(g)
    eps = (aa - 1) / (ndat * bb + aa - 1)
    sel = (uniform(size=aa.shape) < eps) * 1
    aaa = np.vstack((aa, aa - 1)).T[np.arange(aa.shape[0]), sel]
    return gamma(shape=aaa, scale=1 / bb)

    # S   = np.empty((setup.ntemps, setup.nclustmax, setup.p, setup.p))
    # S[:] = np.eye(setup.p) * 1e-4
    # nS  = np.zeros((setup.ntemps, setup.nclustmax))
    # mS  = np.zeros((setup.ntemps, setup.nclustmax, setup.p))
    # cov = [np.empty((setup.ntemps, setup.ns2[i], setup.p, setup.p)) for i in range(setup.nexp)]
    # mu  = [np.empty((setup.ntemps, setup.ns2[i], setup.p)) for i in range(setup.nexp)]


class AMcov_clust:
    """
    We want to do adaptive metropolis (AM) for theta. There are nclustmax theta vectors, but we
    can't do AM on them because they will be switching labels. Instead we do AM on the experiment
    specific theta histories, but then need to combine them to get a sample for the
    cluster. Within each cluster, we weight the means and covariances from experiments within
    the cluster. The way Peter has done this is to accumulate with the proper weights. He
    loops over experiments, which each have a cov and mu, and adds them (weighted) to the
    appropriate cluster cov and mu.

    Adaptive Metropolis for clustered calibration. Notes:
    - You cannot just do AM on 'theta' beause of label switching
    - Instead do AM on 'theta_hist' because it always exists with same dimension
    - Tricky part is that we are then proposing an update to theta instead of theta_hist
    - Combine relevant theta_hists (appropriately weighted) for updating theta
    - https://www.mathworks.com/help/matlab/import_export/using-mapreduce-to-compute-covariance-and-related-quantities.html
    """

    def __init__(
        self,
        nexp,
        ntheta,
        ntemps,
        p,
        nclustmax,
        start_var=1e-4,
        start_adapt_iter=300,
        tau_start=0.0,
    ):
        self.eps = 1.0e-12
        self.AM_SCALAR = 2.4**2 / p
        self.tau = (
            tau_start * np.ones(ntemps)
        )  # note this is different from hier--not sure how to have a different tau for each experiment
        self.S = np.empty((ntemps, nclustmax, p, p))  # proposal covariance
        for t in range(ntemps):
            for c in range(nclustmax):
                self.S[t, c] = np.eye(p) * start_var
        self.cov_hist = [
            np.empty((ntemps, ntheta[i], p, p)) for i in range(nexp)
        ]  # covariance of past sample by experiment
        self.mu_hist = [
            np.empty((ntemps, ntheta[i], p)) for i in range(nexp)
        ]  # mean of past samples by experiment
        self.cov_clust = np.empty((
            ntemps,
            nclustmax,
            p,
            p,
        ))  # covariance of past samples by cluster (weighted combination of cov_hist)
        self.mu_clust = np.zeros((
            ntemps,
            nclustmax,
            p,
        ))  # mean of past sample by cluster (weighted combination of mu_hist)
        self.n_hc = np.zeros((
            ntemps,
            nclustmax,
        ))  # count for accumulation purposes
        self.nexp = nexp
        self.ntheta = ntheta
        self.ntemps = ntemps
        self.nclustmax = nclustmax
        self.p = p
        self.start_adapt_iter = start_adapt_iter
        self.count_100 = np.zeros(self.ntemps, dtype=int)
        self.nS = np.zeros((ntemps, nclustmax))
        self.temps = np.arange(ntemps)

    def update(self, x, m, curr_delta):
        if m > self.start_adapt_iter:
            # update experiment-specific mu and cov
            for i in range(self.nexp):
                self.mu_hist[i] += (x[i][m - 1] - self.mu_hist[i]) / m
                self.cov_hist[i][:] = +((m - 1) / m) * self.cov_hist[i] + (
                    (m - 1) / (m * m)
                ) * np.einsum(
                    "tej,tel->tejl",
                    x[i][m - 1] - self.mu_hist[i],
                    x[i][m - 1] - self.mu_hist[i],
                )
            # reweight experiment-specific to get cluster-specific mu and cov
            self.cluster_covariance_update(m, curr_delta)
            # make history covariance into proposal covariance
            self.S = self.AM_SCALAR * np.einsum(
                "tmjl,t->tmjl",
                self.cov_clust + np.eye(self.p) * self.eps,
                np.exp(self.tau),
            )  # is identity be added correctly (correctly broadcast)?

        elif m == self.start_adapt_iter:
            # calculate experiment-specific mu and cov
            for i in range(self.nexp):
                self.mu_hist[i][:] = x[i][:m].mean(axis=0)
                self.cov_hist[i][:] = cov_4d_pcm(x[i][:m], self.mu_hist[i])

            # reweight experiment-specific to get cluster-specific mu and cov
            self.cluster_covariance_update(m, curr_delta)
            # make history covariance into proposal covariance
            self.S = self.AM_SCALAR * np.einsum(
                "tmjl,t->tmjl",
                self.cov_clust + np.eye(self.p) * self.eps,
                np.exp(self.tau),
            )  # is identity be added correctly (correctly broadcast)?

    def cluster_covariance_update(self, n, delta):
        # https://www.mathworks.com/help/matlab/import_export/using-mapreduce-to-compute-covariance-and-related-quantities.html
        self.cov_clust[:] = (
            0.0  # accumulating covariance                                                                                   #(c1)
        )
        self.mu_clust[:] = (
            0.0  # accumulating mean                                                                                           #(m1)
        )
        self.n_hc[:] = (
            0.0  # accumulating count                                                                                              #(n1)
        )
        mu_new = np.empty((self.ntemps, self.cov_clust.shape[-1]))  # (m)
        n_new = np.zeros((self.ntemps, 1))  # (n)
        for i in range(self.nexp):  # loop over unvectorized experiments
            for j in range(
                delta[i].shape[1]
            ):  # loop over vectorized experiments
                n_new[:] = (
                    self.n_hc[self.temps, delta[i].T[j], None] + n
                )  # (n=n1+n2)
                mu_new[:] = (
                    self.n_hc[self.temps, delta[i].T[j], None]
                    * self.mu_clust[self.temps, delta[i].T[j]]
                    + n * self.mu_hist[i][self.temps, j]
                ) / n_new  # (m=(n1*m1 + n2*m2)/n)
                self.cov_clust[self.temps, delta[i].T[j]] = (
                    1
                    / n_new[:, :, None]
                    * (  # 1/n *
                        +self.n_hc[self.temps, delta[i].T[j], None, None]
                        * self.cov_clust[
                            self.temps, delta[i].T[j]
                        ]  # cov1 (what has accumulated so far)  #(c1)
                        + n
                        * self.cov_hist[i][
                            self.temps, j
                        ]  # cov2 (history for experiment ij)          #(c2)
                        + np.einsum(
                            "t,tp,tq->tpq",
                            self.n_hc[self.temps, delta[i].T[j]],  # mu1
                            self.mu_clust[self.temps, delta[i].T[j]] - mu_new,
                            self.mu_clust[self.temps, delta[i].T[j]] - mu_new,
                        )
                        + n
                        * np.einsum(
                            "tp,tq->tpq",
                            self.mu_hist[i][self.temps, j] - mu_new,
                            self.mu_hist[i][self.temps, j] - mu_new,
                        )  # mu2
                    )
                )
                self.mu_clust[self.temps, delta[i].T[j]] = mu_new  # (m1=m)
                self.n_hc[self.temps, delta[i].T[j]] = n_new.ravel()  # (n1=n)
        return

    def update_tau(
        self, m
    ):  # label switching makes it difficult to track acceptance rate, so this is turned off
        # diminishing adaptation based on acceptance rate for each temperature
        if False:
            if (m % 100 == 0) and (m > 300):
                ddelta = min(0.1, 5 / sqrt(m + 1))
                self.tau[np.where(self.count_100 < 23)] = (
                    self.tau[np.where(self.count_100 < 23)] - ddelta
                )
                self.tau[np.where(self.count_100 > 23)] = (
                    self.tau[np.where(self.count_100 > 23)] + ddelta
                )
                self.count_100 *= 0

    def gen_cand(self, x, m):  # one sample per cluster
        # x_cand = [chol_sample_1per(x[i][m-1], self.S) for i in range(self.nexp)]
        x_cand = chol_sample_1per(x[m - 1], self.S)
        return x_cand


OutCalibClust = namedtuple(
    "OutCalibClust",
    "theta theta_hist s2 count count_temper pred_curr theta0 Sigma0 delta eta nclustmax theta_am",
)

## DP Cluster Calibration


def calibClust(setup, parallel=False):
    t0 = time.time()

    if parallel:
        pool = mp.Pool(processes=mp.cpu_count())
    else:
        pool = None

    ## Constants Declaration

    # boolean with dimension [i] x [ylens[i], ns2[i]] indicating which s2 to use for each output
    s2_ind_mat = [
        (setup.s2_ind[i][:, None] == range(setup.ns2[i]))
        for i in range(setup.nexp)
    ]

    # boolean with dimension [i] x [ylens[i], ntheta[i]] indicating which theta to use for each output
    theta_ind_mat = [
        (setup.theta_ind[i][:, None] == range(setup.ntheta[i]))
        for i in range(setup.nexp)
    ]

    # sequence of length ntemps
    temps = np.arange(setup.ntemps)

    # matrix of repeated inverse temperatures of dimension [ntemps, nclustmax]
    itl_mat_theta = (np.ones((setup.ntemps, setup.nclustmax)).T * setup.itl).T

    # repeated inverse temperatures of dimension [i] x [ntemps, ns2[i]]
    itl_mat_s2 = [
        np.ones((setup.ntemps, setup.ns2[i])) * setup.itl.reshape(-1, 1)
        for i in range(setup.nexp)
    ]

    ## Parameter Declaration
    theta0 = np.empty([setup.nmcmc, setup.ntemps, setup.p])
    # theta0[0] = chol_sample_1per_constraints(
    #        np.zeros((setup.ntemps, setup.p)), np.array([np.eye(setup.p)] * setup.ntemps),
    #        setup.checkConstraints, setup.bounds_mat, setup.bounds.keys(), setup.bounds, setup.constants
    #        )
    theta0_start = initfunc_unif(size=[setup.ntemps, setup.p])
    good = setup.checkConstraints(
        tran_unif(theta0_start, setup.bounds_mat, setup.bounds.keys()),
        setup.bounds,
    )
    while np.any(np.logical_not(good)):
        theta0_start[np.where(np.logical_not(good))] = initfunc_unif(
            size=[(np.logical_not(good)).sum(), setup.p]
        )
        good[np.where(np.logical_not(good))] = setup.checkConstraints(
            tran_unif(
                theta0_start[np.where(np.logical_not(good))],
                setup.bounds_mat,
                setup.bounds.keys(),
            ),
            setup.bounds,
        )
    theta0[0] = theta0_start

    Sigma0 = np.empty([setup.nmcmc, setup.ntemps, setup.p, setup.p])
    Sigma0[0] = np.eye(setup.p) * 0.25**2
    # initialize delta, the cluster membership indicator (for each experiment)
    delta = [
        np.empty([setup.nmcmc, setup.ntemps, setup.ntheta[i]], dtype=int)
        for i in range(setup.nexp)
    ]
    for i in range(setup.nexp):
        delta[i][0] = choice(
            setup.nclustmax, size=(setup.ntemps, setup.ntheta[i])
        )
    delta_ind_mat = [
        (delta[i][0, :, :, None] == range(setup.nclustmax))
        for i in range(setup.nexp)
    ]
    clust_mem_count = np.zeros(
        (setup.ntemps, setup.nclustmax), dtype=int
    )  # (extant) cluster weight for ijth experiment
    clust_nomem_bool = np.zeros(clust_mem_count.shape, dtype=bool)
    # clust_nomem_bool = np.zeros(delta_ind_mat.shape, dtype = bool)                  # (njs > 0) -> false
    djs = np.empty(
        clust_mem_count.shape
    )  # (candidate) cluster weight for ijth experiment
    cluster_weights = np.empty(clust_mem_count.shape)
    for i in range(setup.nexp):
        clust_mem_count[:] += bincount2D_vectorized(
            delta[i][0], setup.nclustmax
        )
    curr_delta = [delta[i][0].copy() for i in range(setup.nexp)]
    theta_unravel = [
        np.repeat(np.arange(setup.ntemps), setup.ntheta[i])
        for i in range(setup.nexp)
    ]
    # initialize Theta
    theta_long_shape = (setup.ntemps * setup.nclustmax, setup.p)
    theta_wide_shape = (setup.ntemps, setup.nclustmax, setup.p)
    theta = np.empty((setup.nmcmc, setup.ntemps, setup.nclustmax, setup.p))
    theta[0] = chol_sample_nper_constraints(
        theta0[0],
        Sigma0[0],
        setup.nclustmax,
        setup.checkConstraints,
        setup.bounds_mat,
        setup.bounds.keys(),
        setup.bounds,
        setup.constants,
    )
    theta_hist = [
        np.empty((setup.nmcmc, setup.ntemps, setup.ntheta[i], setup.p))
        for i in range(setup.nexp)
    ]
    theta_cand = np.empty(theta_wide_shape)
    for i in range(setup.nexp):
        theta_hist[i][0] = theta[
            0, theta_unravel[i], delta[i][0].ravel()
        ].reshape(setup.ntemps, setup.ns2[i], setup.p)
    theta_eval = np.empty(theta_wide_shape)
    theta_ext = np.zeros(
        (setup.ntemps, setup.nclustmax), dtype=bool
    )  # array of (current) extant theta locs
    for i in range(setup.nexp):
        theta_ext[theta_unravel[i], delta[i][0].ravel()] += True
    ntheta = theta_ext.sum(axis=1)  # count extant thetas
    # initialize sigma2 and eta
    log_s2 = [
        np.ones([setup.nmcmc, setup.ntemps, setup.ns2[i]])
        for i in range(setup.nexp)
    ]
    eta = np.empty((setup.nmcmc, setup.ntemps))
    eta[0] = 5.0

    s2_which_mat = [
        [np.where(s2_ind_mat[i][:, j])[0] for j in range(setup.ntheta[i])]
        for i in range(setup.nexp)
    ]
    theta_which_mat = [
        [np.where(theta_ind_mat[i][:, j])[0] for j in range(setup.ntheta[i])]
        for i in range(setup.nexp)
    ]

    ## Initialize *Current* Variables

    # LAUREN COMMENTED OUT LAST THREE ELEMENTS
    # pred_curr       = [None] * setup.nexp # [i] x [ntemps, ylens[i]]
    # llik_curr        = [None] * setup.nexp # [i], ntheta[i] x ntemps ### was [i] x [ntemps, ns2[i]]
    # llik_cand  = [None] * setup.nexp # [i], ntheta[i] x ntemps
    # dev_sq          = [None] * setup.nexp # [i] x [ntemps, ns2[i]]
    pred_curr_delta = [None] * setup.nexp  # [i] x [ntemps, nclustmax, ylens[i]]
    llik_curr_delta = [None] * setup.nexp  # [i] x [ntemps, nclustmax, ns2[i]]
    pred_cand_theta = [None] * setup.nexp  # [i] x [ntemps, ylens[i]]
    pred_curr_theta = [None] * setup.nexp  # [i] x [ntemps, ylens[i]]
    llik_cand_theta = [None] * setup.nexp  # [i] x [ntemps, ns2[i]]
    llik_curr_theta = [None] * setup.nexp  # [i] x [ntemps, ns2[i]]
    llik_cand_theta_ = np.zeros((
        setup.ntemps,
        setup.nclustmax,
    ))  # above, Summarized by cluster assignment (i.e., llik for each unique theta, summed over experiments that share that theta)
    llik_curr_theta_ = np.zeros((setup.ntemps, setup.nclustmax))  #
    marg_lik_cov_curr = (
        [None] * setup.nexp
    )  # does not need to get summed over experiments (should be one for each experiment, regardless of cluster)
    for i in range(setup.nexp):
        ### Initialize predictions for theta_i's
        pred_curr_theta[i] = setup.models[i].eval(
            tran_unif(
                theta_hist[i][0].reshape(-1, setup.p),  # 0 for first iteration
                setup.bounds_mat,
                setup.bounds.keys(),
            ),
            False,
        )
        pred_cand_theta[i] = pred_curr_theta[i].copy()
        marg_lik_cov_curr[i] = [None] * setup.ntemps
        llik_curr_theta[i] = np.empty([setup.ntemps, setup.ntheta[i]])
        for t in range(setup.ntemps):
            marg_lik_cov_curr[i][t] = [None] * setup.ntheta[i]
            s2_stretched = log_s2[i][0][t, setup.theta_ind[i]]
            for j in range(setup.ntheta[i]):
                marg_lik_cov_curr[i][t][j] = setup.models[i].lik_cov_inv(
                    np.exp(s2_stretched[s2_which_mat[i][j]])
                )
                llik_curr_theta[i][t][j] = setup.models[i].llik(
                    setup.ys[i][theta_which_mat[i][j]],
                    pred_curr_theta[i][t][theta_which_mat[i][j]],
                    marg_lik_cov_curr[i][t][j],
                )
        llik_cand_theta[i] = llik_curr_theta[i].copy()

        ### Initialize predictions for clusters
        pred_curr_delta[i] = (
            setup.models[i]
            .eval(
                tran_unif(
                    theta[0].reshape(-1, setup.p),
                    setup.bounds_mat,
                    setup.bounds.keys(),
                ),
                True,
            )
            .reshape(setup.ntemps, setup.nclustmax, setup.y_lens[i])
        )
        llik_curr_delta[i] = np.empty([
            setup.ntemps,
            setup.nclustmax,
            setup.ntheta[i],
        ])
        for t in range(setup.ntemps):
            s2_stretched = log_s2[i][0][t, setup.theta_ind[i]]
            for j in range(setup.ntheta[i]):
                for k in range(setup.nclustmax):
                    llik_curr_delta[i][t][k][j] = setup.models[i].llik(
                        setup.ys[i][theta_which_mat[i][j]],
                        pred_curr_delta[i][t][k][theta_which_mat[i][j]],
                        marg_lik_cov_curr[i][t][j],
                    )

    ## Initialize Adaptive Metropolis related Variables
    # S   = np.empty((setup.ntemps, setup.nclustmax, setup.p, setup.p))
    # S[:] = np.eye(setup.p) * 1e-4
    # nS  = np.zeros((setup.ntemps, setup.nclustmax))
    # mS  = np.zeros((setup.ntemps, setup.nclustmax, setup.p))
    # cov = [np.empty((setup.ntemps, setup.ns2[i], setup.p, setup.p)) for i in range(setup.nexp)]
    # mu  = [np.empty((setup.ntemps, setup.ns2[i], setup.p)) for i in range(setup.nexp)]
    cov_theta_cand = AMcov_clust(
        nexp=setup.nexp,
        ntheta=np.array([setup.ntheta[i] for i in range(setup.nexp)]),
        ntemps=setup.ntemps,
        p=setup.p,
        nclustmax=setup.nclustmax,
        start_var=setup.start_var_theta,
        start_adapt_iter=setup.start_adapt_iter,
        tau_start=setup.start_tau_theta,
    )

    ## Initialize Theta0 and Sigma0 related variables
    theta0_prior_cov = (
        setup.theta0_prior_cov
    )  # was 10^2, but that is far from uniform when back transforming
    theta0_prior_prec = scipy.linalg.inv(theta0_prior_cov)
    theta0_prior_mean = setup.theta0_prior_mean  # np.repeat(0.5, setup.p)
    theta0_prior_ldet = slogdet(theta0_prior_cov)[1]
    tbar = np.empty(theta0[0].shape)
    mat = np.zeros((setup.ntemps, setup.p, setup.p))
    Sigma0_prior_df = setup.Sigma0_prior_df  # setup.p
    Sigma0_prior_scale = (
        setup.Sigma0_prior_scale
    )  # np.eye(setup.p)*.1**2#/setup.p
    Sigma0_dfs = Sigma0_prior_df + setup.nclustmax * setup.itl
    Sigma0_scales = np.empty((setup.ntemps, setup.p, setup.p))
    Sigma0_scales[:] = Sigma0_prior_scale
    Sigma0_ldet_curr = slogdet(Sigma0[0])[1]
    Sigma0_inv_curr = np.linalg.inv(Sigma0[0])
    cc = np.empty((setup.ntemps, setup.p, setup.p))
    dd = np.empty((setup.ntemps, setup.p))
    ## Initialize Counters
    count_temper = np.zeros([setup.ntemps, setup.ntemps])
    count = np.zeros(setup.ntemps)
    ## Initialize Alphas
    alpha_long_shape = (setup.ntemps * setup.nclustmax,)
    alpha_wide_shape = (
        setup.ntemps,
        setup.nclustmax,
    )
    alpha = np.ones(alpha_wide_shape) * -np.inf
    accept = np.zeros(alpha_wide_shape, dtype=bool)
    sw_alpha = np.zeros(setup.nswap_per)
    good_values = np.zeros(alpha_wide_shape, dtype=bool)

    delta_size = [
        (setup.ntemps, setup.nclustmax, setup.ns2[i]) for i in range(setup.nexp)
    ]
    cluster_cum_prob = np.empty((setup.ntemps, setup.nclustmax))
    cluster_sample_unif = [
        np.empty((setup.ntemps, setup.ns2[i])) for i in range(setup.nexp)
    ]

    ## start MCMC
    for m in range(1, setup.nmcmc):
        # ------------------------------------------------------------------------------------------
        ## Gibbs Update for delta (cluster identifier)
        for i in range(setup.nexp):
            delta[i][m] = delta[i][m - 1]

        #   Establish initial cluster weighting
        clust_mem_count[:] = 0.0
        for l in range(setup.nexp):
            clust_mem_count[:] += bincount2D_vectorized(
                delta[l][m], setup.nclustmax
            )
        clust_nomem_bool[:] = (
            clust_mem_count == 0
        )  # fixed at start of iteration
        # if (non-extant) cluster j becomes extant, then set this to False.

        ##################
        ### Draw Delta ###
        ##################
        for i in range(setup.nexp):
            cluster_sample_unif[i][:] = uniform(
                size=(setup.ntemps, setup.ns2[i])
            )
            for j in range(setup.ntheta[i]):
                # weighting assigned to extant clusters for ij'th within-vectorized experiment
                clust_mem_count[temps, delta[i][m, :, j]] -= 1
                # weighting assigned to candidate (non-extant) clusters
                djs[:] = clust_nomem_bool * (
                    eta[m - 1] / (clust_nomem_bool.sum(axis=1) + 1e-9)
                ).reshape(-1, 1)
                # djs[:] = (njs == 0) * (eta[m-1] / (njs == 0).sum(axis = 1)).reshape(-1,1)
                # unnormalized log-probability of cluster membership
                with np.errstate(divide="ignore", invalid="ignore"):
                    cluster_cum_prob[:] = (
                        +np.log(
                            clust_mem_count + djs
                        )  # this line doesnt match paper? ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        + llik_curr_delta[i][:, :, j] * setup.itl.reshape(-1, 1)
                    )
                # np.nan_to_num(cluster_cum_prob, False, nan = -np.inf, neginf = -np.inf)
                # ---fix for numerical stability in np.exp
                cluster_cum_prob -= cluster_cum_prob.max(axis=1).reshape(-1, 1)
                # Un-normalized cumulative probability of cluster membership
                cluster_cum_prob[:] = np.exp(cluster_cum_prob).cumsum(axis=1)
                # Normalized cumulative probability of cluster membership
                cluster_cum_prob /= cluster_cum_prob[:, -1].reshape(-1, 1)
                delta[i][m, :, j] = (
                    cluster_sample_unif[i][:, j].reshape(-1, 1)
                    > cluster_cum_prob
                ).sum(axis=1)
                # Fix weights using new cluster assignments for ij'th within-vectorized experiment
                clust_mem_count[temps, delta[i][m, :, j]] += 1
                # If a candidate cluster became extant--remove candidate flag.
                clust_nomem_bool[clust_mem_count > 0] = False

            delta_ind_mat[i][:] = delta[i][m, :, :, None] == range(
                setup.nclustmax
            )
            curr_delta[i][:] = delta[i][m]

        ##########################
        ### Update Theta Evals ###
        ##########################
        for i in range(setup.nexp):
            pred_curr_theta[i][:] = setup.models[i].eval(
                tran_unif(
                    theta[m - 1, theta_unravel[i], delta[i][m].ravel()],
                    setup.bounds_mat,
                    setup.bounds.keys(),
                )
            )  # update after delta update before
            for t in range(setup.ntemps):
                for j in range(setup.ntheta[i]):
                    llik_curr_theta[i][t][j] = setup.models[i].llik(
                        setup.ys[i][theta_which_mat[i][j]],
                        pred_curr_theta[i][t][theta_which_mat[i][j]],
                        marg_lik_cov_curr[i][t][j],
                    )

        # ------------------------------------------------------------------------------------------
        ## adaptive Metropolis per Cluster
        if False:
            if m > 300:
                for i in range(setup.nexp):
                    mu[i] += (theta_hist[i][m - 1] - mu[i]) / m
                    cov[i][:] = +((m - 1) / m) * cov[i] + (
                        (m - 1) / (m * m)
                    ) * np.einsum(
                        "tej,tel->tejl",
                        theta_hist[i][m - 1] - mu[i],
                        theta_hist[i][m - 1] - mu[i],
                    )
                cluster_covariance_update(
                    S,
                    mS,
                    nS,
                    m,
                    curr_delta,
                    cov,
                    mu,
                    setup.nexp,
                    setup.nclustmax,
                    temps,
                )
                S = AM_SCALAR * np.einsum("ijkl,i->ijkl", S, np.exp(tau))
            elif m == 300:
                for i in range(setup.nexp):
                    mu[i][:] = theta_hist[i][:m].mean(axis=0)
                    cov[i][:] = cov_4d_pcm(theta_hist[i][:m], mu[i])
                cluster_covariance_update(
                    S,
                    mS,
                    nS,
                    m,
                    curr_delta,
                    cov,
                    mu,
                    setup.nexp,
                    setup.nclustmax,
                    temps,
                )
                S = AM_SCALAR * np.einsum("ijkl,i->ijkl", S, np.exp(tau))
            else:
                pass
        # ------------------------------------------------------------------------------------------

        ###################
        ### Draw Thetas ###
        ###################

        llik_cand_theta_[:] = 0.0
        llik_curr_theta_[:] = 0.0
        theta[m] = theta[m - 1]
        theta_eval[:] = theta[m - 1]

        cov_theta_cand.update(theta_hist, m, curr_delta)
        theta_cand = cov_theta_cand.gen_cand(theta, m)
        good_values[:] = setup.checkConstraints(
            tran_unif(
                theta_cand.reshape(theta_long_shape),
                setup.bounds_mat,
                setup.bounds.keys(),
            ),
            setup.bounds,
        ).reshape(setup.ntemps, setup.nclustmax)
        theta_eval[good_values] = theta_cand[good_values]
        for i in range(setup.nexp):
            pred_cand_theta[i][:] = setup.models[i].eval(
                tran_unif(
                    theta_eval[theta_unravel[i], delta[i][m].ravel()],
                    setup.bounds_mat,
                    setup.bounds.keys(),
                )
            )
            for t in range(setup.ntemps):
                for j in range(setup.ntheta[i]):
                    llik_cand_theta[i][t][j] = setup.models[i].llik(
                        setup.ys[i][theta_which_mat[i][j]],
                        pred_cand_theta[i][t][theta_which_mat[i][j]],
                        marg_lik_cov_curr[i][t][j],
                    )
                    llik_cand_theta_[t][curr_delta[i][t, j]] += llik_cand_theta[
                        i
                    ][t][j]
                    llik_curr_theta_[t][curr_delta[i][t, j]] += llik_curr_theta[
                        i
                    ][t][j]

        alpha[:] = -np.inf
        alpha[good_values] = (
            itl_mat_theta * (llik_cand_theta_ - llik_curr_theta_)
            + mvnorm_logpdf_(
                theta_cand, theta0[m - 1], Sigma0_inv_curr, Sigma0_ldet_curr
            )
            - mvnorm_logpdf_(
                theta[m - 1], theta0[m - 1], Sigma0_inv_curr, Sigma0_ldet_curr
            )
        )[good_values]
        accept[:] = np.log(uniform(size=alpha.shape)) < alpha
        theta[m, accept] = theta_cand[accept]

        for i in range(setup.nexp):
            theta_hist[i][m] = theta[
                m, theta_unravel[i], delta[i][m].ravel()
            ].reshape(setup.ntemps, setup.ns2[i], setup.p)
            pred_curr_theta[i][:] = setup.models[i].eval(
                tran_unif(
                    theta_hist[i][m].reshape(-1, setup.p),
                    setup.bounds_mat,
                    setup.bounds.keys(),
                ),
                False,
            )
            llik_curr_theta[i][:] = llik_cand_theta[i][:]

        count += accept.sum(axis=1)
        cov_theta_cand.count_100 += accept.sum(axis=1)

        cov_theta_cand.update_tau(m)
        theta_ext[:] = False
        for i in range(setup.nexp):
            theta_ext[theta_unravel[i], delta[i][m].ravel()] += True
        ntheta[:] = theta_ext.sum(axis=1)

        ########################
        ###  Gibbs update s2 ###
        ########################
        for i in range(setup.nexp):
            if setup.models[i].s2 == "gibbs":
                dev_sq = (pred_curr_theta[i] - setup.ys[i]) ** 2 @ s2_ind_mat[
                    i
                ]  # squared deviations
                log_s2[i][m] = np.log(
                    1
                    / np.random.gamma(
                        itl_mat_s2[i] * (setup.ny_s2[i] / 2 + setup.ig_a[i] + 1)
                        - 1,
                        1 / (itl_mat_s2[i] * (setup.ig_b[i] + dev_sq / 2)),
                    )
                )
                for t in range(setup.ntemps):
                    s2_stretched = log_s2[i][m][t, setup.theta_ind[i]]
                    for j in range(setup.ntheta[i]):
                        marg_lik_cov_curr[i][t][j] = setup.models[
                            i
                        ].lik_cov_inv(np.exp(s2_stretched[s2_which_mat[i][j]]))
                        llik_curr_theta[i][t][j] = setup.models[i].llik(
                            setup.ys[i][s2_which_mat[i][j]],
                            pred_curr_theta[i][t][s2_which_mat[i][j]],
                            marg_lik_cov_curr[i][t][j],
                        )

            elif setup.models[i].s2 == "fix":
                log_s2[i][m] = np.log(setup.sd_est[i] ** 2)

                for t in range(setup.ntemps):
                    s2_stretched = log_s2[i][m][t, setup.theta_ind[i]]
                    for j in range(setup.ntheta[i]):
                        marg_lik_cov_curr[i][t][j] = setup.models[
                            i
                        ].lik_cov_inv(np.exp(s2_stretched[s2_which_mat[i][j]]))
                        llik_curr_theta[i][t][j] = setup.models[i].llik(
                            setup.ys[i][s2_which_mat[i][j]],
                            pred_curr_theta[i][t][s2_which_mat[i][j]],
                            marg_lik_cov_curr[i][t][j],
                        )

            else:
                print("TODO: fill in s2 sampling without gibbs")

        ##########################
        ### Update Delta Evals ### (THIS IS THE SLOW CHUNK THAT IS MESSING WITH RUNTIME!!!! Problem is compute_state_history looping over NHist)
        ##########################
        for i in range(setup.nexp):
            pred_curr_delta[i][:] = (
                setup.models[i]
                .eval(  # (IN PARTICULAR, THIS CALL!!!!)
                    tran_unif(
                        theta[m]
                        .reshape(-1, setup.p)
                        .repeat(setup.ns2[i], axis=0),
                        setup.bounds_mat,
                        setup.bounds.keys(),
                    ),
                )
                .reshape(setup.ntemps, setup.nclustmax, setup.y_lens[i])
            )
            for t in range(setup.ntemps):
                for j in range(setup.ntheta[i]):
                    for k in range(setup.nclustmax):
                        llik_curr_delta[i][t][k][j] = setup.models[i].llik(
                            setup.ys[i][theta_which_mat[i][j]],
                            pred_curr_delta[i][t][k][theta_which_mat[i][j]],
                            marg_lik_cov_curr[i][t][j],
                        )

        ###########################
        ### Gibbs Update Theta0 ###
        ###########################
        cc[:] = np.linalg.inv(
            np.einsum("t,tpq->tpq", ntheta * setup.itl, Sigma0_inv_curr)
            + theta0_prior_prec,
        )
        tbar[:] = (
            theta[m] * theta_ext.reshape(setup.ntemps, setup.nclustmax, 1)
        ).sum(axis=1)
        tbar[:] /= ntheta.reshape(setup.ntemps, 1)
        dd[:] = +np.einsum(
            "t,tl->tl",
            setup.itl,
            np.einsum("t,tlk,tk->tl", ntheta, Sigma0_inv_curr, tbar),
        ) + np.dot(theta0_prior_prec, theta0_prior_mean)
        theta0[m][:] = chol_sample_1per_constraints(
            np.einsum("tlk,tk->tl", cc, dd),
            cc,
            setup.checkConstraints,
            setup.bounds_mat,
            setup.bounds.keys(),
            setup.bounds,
            setup.constants,
        )

        ###########################
        ### Gibbs update Sigma0 ###
        ###########################
        mat[:] *= 0.0
        # for i in range(setup.nexp): # This loop appears to be extra (just inflating) - devin
        # mat3 = mat*0
        # mat3 += np.einsum(
        #        'tnp,tnq->tpq',
        #        (theta[m] - theta0[m].reshape(setup.ntemps, 1, setup.p))* theta_ext.reshape(setup.ntemps, setup.nclustmax, 1),
        #        (theta[m] - theta0[m].reshape(setup.ntemps, 1, setup.p))* theta_ext.reshape(setup.ntemps, setup.nclustmax, 1),
        #        )
        mat += np.einsum(
            "t,tnp,tnq->tpq",
            setup.itl,
            (
                (theta[m] - theta0[m].reshape(setup.ntemps, 1, setup.p))
                * theta_ext.reshape(setup.ntemps, setup.nclustmax, 1)
            ),
            (
                (theta[m] - theta0[m].reshape(setup.ntemps, 1, setup.p))
                * theta_ext.reshape(setup.ntemps, setup.nclustmax, 1)
            ),
        )
        # mat2 = mat*0.
        # for ii in np.where(theta_ext.reshape(setup.ntemps, setup.nclustmax, 1))[1]:
        #    mat2 += (theta[m,0,ii,:]-theta0[m,0,:]).reshape(setup.p,1) @ (theta[m,0,ii,:]-theta0[m,0,:]).reshape(setup.p,1).T
        # theta[m,0,np.where(theta_ext.reshape(setup.ntemps, setup.nclustmax, 1))[1],:].T @ theta[m,0,np.where(theta_ext.reshape(setup.ntemps, setup.nclustmax, 1))[1],:]
        Sigma0_scales[:] = Sigma0_prior_scale + mat
        Sigma0_dfs[:] = Sigma0_prior_df + theta_ext.sum(axis=1) * setup.itl
        for t in range(setup.ntemps):
            Sigma0[m, t] = invwishart.rvs(
                df=Sigma0_dfs[t], scale=Sigma0_scales[t]
            )
        Sigma0_ldet_curr[:] = np.linalg.slogdet(Sigma0[m])[1]
        Sigma0_inv_curr[:] = np.linalg.inv(Sigma0[m])

        ###############################################
        ### Gibbs Update for Theta (Not in Cluster) ###
        ###############################################
        theta_cand[:] = chol_sample_nper_constraints(
            theta0[m],
            Sigma0[m],
            setup.nclustmax,
            setup.checkConstraints,
            setup.bounds_mat,
            setup.bounds.keys(),
            setup.bounds,
            setup.constants,
        )
        theta[m, ~theta_ext] = theta_cand[~theta_ext]

        # TODO: add decorrelation step here?

        ########################
        ### Gibbs Update eta ###
        ########################
        eta[m] = sample_eta(
            eta[m - 1],
            ntheta,
            sum(setup.ns2[i] for i in range(setup.nexp)),
            setup.eta_prior_shape,
            setup.eta_prior_rate,
        )

        #######################
        ### Tempering Swaps ###
        #######################
        if m > setup.start_temper and setup.ntemps > 1:
            for _ in range(setup.nswap):
                sw = np.random.choice(
                    setup.ntemps, 2 * setup.nswap_per, replace=False
                ).reshape(-1, 2)
                sw_alpha[:] = 0.0  # reset swap probability
                sw_alpha[:] = sw_alpha + (
                    setup.itl[sw.T[1]] - setup.itl[sw.T[0]]
                ) * (
                    -mvnorm_logpdf(
                        theta0[m][sw.T[1]],
                        theta0_prior_mean,
                        theta0_prior_prec,
                        theta0_prior_ldet,
                    )
                    + mvnorm_logpdf(
                        theta0[m][sw.T[0]],
                        theta0_prior_mean,
                        theta0_prior_prec,
                        theta0_prior_ldet,
                    )
                    - invwishart_logpdf(
                        Sigma0[m][sw.T[1]], Sigma0_prior_df, Sigma0_prior_scale
                    )
                    + invwishart_logpdf(
                        Sigma0[m][sw.T[0]], Sigma0_prior_df, Sigma0_prior_scale
                    )
                    - gamma_logpdf(eta[m][sw.T[1]], 2, 0.1)
                    + gamma_logpdf(eta[m][sw.T[0]], 2, 0.1)
                )
                for i in range(setup.nexp):
                    sw_alpha[:] = sw_alpha + (
                        setup.itl[sw.T[1]] - setup.itl[sw.T[0]]
                    ) * (
                        # for t_0
                        +setup.s2_prior_kern[i](
                            np.exp(log_s2[i][m][sw.T[0]]),
                            setup.ig_a[i],
                            setup.ig_b[i],
                        ).sum(axis=1)
                        + (
                            mvnorm_logpdf_(
                                theta[m, sw.T[0]],
                                theta0[m, sw.T[0]],
                                Sigma0_inv_curr[sw.T[0]],
                                Sigma0_ldet_curr[sw.T[0]],
                            )
                            * theta_ext[sw.T[0]]
                        ).sum(axis=1)
                        # - 0.5 * (setup.ny_s2[i] * np.log(s2[i][m])).sum(axis = 1)[sw.T[0]]
                        # - 0.5 * sse_curr[i][sw.T[0]].sum(axis = 1)
                        + llik_curr_theta[i][sw.T[0]].sum(axis=1)
                        # for t_1
                        - setup.s2_prior_kern[i](
                            np.exp(log_s2[i][m][sw.T[1]]),
                            setup.ig_a[i],
                            setup.ig_b[i],
                        ).sum(axis=1)
                        - (
                            mvnorm_logpdf_(
                                theta[m, sw.T[1]],
                                theta0[m, sw.T[1]],
                                Sigma0_inv_curr[sw.T[1]],
                                Sigma0_ldet_curr[sw.T[1]],
                            )
                            * theta_ext[sw.T[1]]
                        ).sum(axis=1)
                        # + 0.5 * (setup.ny_s2[i] * np.log(s2[i][m])).sum(axis = 1)[sw.T[1]]
                        # + 0.5 * sse_curr[i][sw.T[1]].sum(axis = 1)
                        - llik_curr_theta[i][sw.T[1]].sum(axis=1)
                    )

                    if False:
                        sw_alpha[:] = sw_alpha + (
                            setup.itl[sw.T[1]] - setup.itl[sw.T[0]]
                        ) * (
                            # for t_0
                            +invgamma_logpdf(
                                s2[i][m][sw.T[0]], setup.ig_a[i], setup.ig_b[i]
                            )
                            + (
                                mvnorm_logpdf_(
                                    theta[m, sw.T[0]],
                                    theta0[m, sw.T[0]],
                                    Sigma0_inv_curr[sw.T[0]],
                                    Sigma0_ldet_curr[sw.T[0]],
                                )
                                * theta_ext[sw.T[0]]
                            ).sum(axis=1)
                            - 0.5
                            * (setup.ny_s2[i] * np.log(s2[i][m])).sum(axis=1)[
                                sw.T[0]
                            ]
                            - 0.5 * sse_curr[i][sw.T[0]].sum(axis=1)
                            # for t_1
                            - invgamma_logpdf(
                                s2[i][m][sw.T[1]], setup.ig_a[i], setup.ig_b[i]
                            )
                            - (
                                mvnorm_logpdf_(
                                    theta[m, sw.T[1]],
                                    theta0[m, sw.T[1]],
                                    Sigma0_inv_curr[sw.T[1]],
                                    Sigma0_ldet_curr[sw.T[1]],
                                )
                                * theta_ext[sw.T[1]]
                            ).sum(axis=1)
                            + 0.5
                            * (setup.ny_s2[i] * np.log(s2[i][m])).sum(axis=1)[
                                sw.T[1]
                            ]
                            + 0.5 * sse_curr[i][sw.T[1]].sum(axis=1)
                        )
                for tt in sw[
                    np.where(
                        sw[np.log(uniform(size=setup.nswap_per)) < sw_alpha]
                    )[0]
                ]:
                    count_temper[tt[0], tt[1]] = count_temper[tt[0], tt[1]] + 1
                    theta[m, tt[0]], theta[m, tt[1]] = (
                        theta[m, tt[1]].copy(),
                        theta[m, tt[0]].copy(),
                    )
                    theta0[m, tt[0]], theta0[m, tt[1]] = (
                        theta0[m, tt[1]].copy(),
                        theta0[m, tt[0]].copy(),
                    )
                    Sigma0[m, tt[0]], Sigma0[m, tt[1]] = (
                        Sigma0[m, tt[1]].copy(),
                        Sigma0[m, tt[0]].copy(),
                    )
                    Sigma0_inv_curr[tt[0]], Sigma0_inv_curr[tt[1]] = (
                        Sigma0_inv_curr[tt[1]].copy(),
                        Sigma0_inv_curr[tt[0]].copy(),
                    )
                    Sigma0_ldet_curr[tt[0]], Sigma0_ldet_curr[tt[1]] = (
                        Sigma0_ldet_curr[tt[1]].copy(),
                        Sigma0_ldet_curr[tt[0]].copy(),
                    )
                    for i in range(setup.nexp):
                        theta_hist[i][m, tt[0]], theta_hist[i][m, tt[1]] = (
                            theta_hist[i][m, tt[1]].copy(),
                            theta_hist[i][m, tt[0]].copy(),
                        )
                        delta[i][m, tt[0]], delta[i][m, tt[1]] = (
                            delta[i][m, tt[1]].copy(),
                            delta[i][m, tt[0]].copy(),
                        )
                        log_s2[i][m, tt[0]], log_s2[i][m, tt[1]] = (
                            log_s2[i][m, tt[1]].copy(),
                            log_s2[i][m, tt[0]].copy(),
                        )
                        pred_curr_theta[i][tt[0]], pred_curr_theta[i][tt[1]] = (
                            pred_curr_theta[i][tt[1]].copy(),
                            pred_curr_theta[i][tt[0]].copy(),
                        )
                        llik_curr_theta[i][tt[0]], llik_curr_theta[i][tt[1]] = (
                            llik_curr_theta[i][tt[1]].copy(),
                            llik_curr_theta[i][tt[0]].copy(),
                        )
                        pred_curr_delta[i][tt[0]], pred_curr_delta[i][tt[1]] = (
                            pred_curr_delta[i][tt[1]].copy(),
                            pred_curr_delta[i][tt[0]].copy(),
                        )
                        llik_curr_delta[i][tt[0]], llik_curr_delta[i][tt[1]] = (
                            llik_curr_delta[i][tt[1]].copy(),
                            llik_curr_delta[i][tt[0]].copy(),
                        )
                    theta_ext[tt[0]], theta_ext[tt[1]] = (
                        theta_ext[tt[1]].copy(),
                        theta_ext[tt[0]].copy(),
                    )
        print(
            "\rCalibration MCMC {:.01%} Complete".format(m / setup.nmcmc),
            end="",
        )
    # ------------------------------------------------------------------------------------------
    pred_curr = pred_curr_theta
    t1 = time.time()
    print("\rCalibration MCMC Complete. Time: {:f} seconds.".format(t1 - t0))
    count_temper = (
        count_temper + count_temper.T - np.diag(np.diag(count_temper))
    )
    out = OutCalibClust(
        theta,
        theta_hist,
        log_s2,
        count,
        count_temper,
        pred_curr,
        theta0,
        Sigma0,
        delta,
        eta,
        setup.nclustmax,
        cov_theta_cand,
    )
    return out


# EOF
